c==== ==================================================================
c
      subroutine dmevas (n,m,ncmplx, A,lda, B,ldb, l, nn, info,
     &                   QG,ldqg, colqg, Gcol,
     &                   QH,ldqh, colqh, Hcol, Fcol,
     &                   tol, Iwork, Rwork, iwarn, ierr)
c
c
c    Purpose
c    =======
c
c    To compute a real matrix F so that the "closed-loop" matrix (A - B*F)
c    has a specified set of eigenvalues.
c
c    Here A and B are real matrices such that the system (B,A) is in "upper
c    staircase" (or "controllability") form, and
c    the set of specified eigenvalues has the property that the complex
c    conjugate of any complex member is also a member.
c
c
c    Argument List
c    =============
c
c    Arguments In
c    ------------
c
c    N      INTEGER
c           Row and column dimension of matrix A
c           Row dimension of matrix B
c           Length of vector of eigenvalues L
c           N .ge. 1
c
c    M      INTEGER
c           Column dimension of matrix B
c           M .ge. 1
c
c    NCMPLX INTEGER
c           (Even) number of complex eigenvalues in L. (See L below)
c           0 .le. NCMPLX .le. N;  NCMPLX even
c
c    A      DOUBLE PRECISION array of DIMENSION (LDA,N)
c           The leading N by N part of this array must contain the state
c           transition matrix A in controllability (upper staircase) form,
c           with the staircase blocks in upper triangular form.
c           Note: this array is overwritten.
c
c    LDA    INTEGER
c           Row dimension of array A, as declared in the calling program
c           LDA .ge. N
c
c    B      DOUBLE PRECISION array of DIMENSION (LDB,M)
c           The leading N by M part of this array must contain the input
c           matrix B in controllability form.
c           Note: this array is overwritten
c
c    LDB    INTEGER
c           Row dimension of array B, as declared in the calling program
c           LDB .ge. N
c
c    L      DOUBLE PRECISION array of DIMENSION (N)
c           Vector of eigenvalues to be allocated.
c           The complex eigenvalues (there are NCMPLX of them) must occur as
c           conjugate pairs.  They are stored in L(1:NCMPLX),  and the
c           real eigenvalues (there are N-NCMPLX of them) are stored in
c           L(NCMPLX+1:N)
c           Since the real and imaginary parts of a complex number 
c           also determine its conjugate, only one real part and one
c           imaginary part are stored for each pair of conjugates. These
c           parts are stored in successive elements of L, with the real
c           parts having odd indices.
c
c           EXAMPLE:
c              To store the four complex eigenvalues
c                     (0.1, 0.2), (0.1, -0.2), (0.3, -0.4), (0.3, 0.4)
c              and the two real eigenvalues
c                     0.5, 0.6
c              L may be initialized to
c                     0.1, 0.2, 0.3, -0.4, 0.5, 0.6
c
c           Observe that for odd i < NCMPLX,  L(i) and L(i+1) are the
c           real and imaginary parts, respectively, of either member of a
c           pair of complex conjugate eigenvalues, as required.
c
c           Note: this array is overwritten. (That is, it may be rearranged).
c
c    NN     INTEGER array of DIMENSION (kp1)
c           where kp1 = (1 + controllability index)  of the system [B,A],  
c           Vector of ranks of B and staircase blocks of A.
c           NN(1) = rank of B
c           NN(k) = rank of (k,k-1) block element of A, for k = 2,...,kp1-1
c           Furthermore, it is important that
c                          NN(kp1) = 0  
c           as the subroutine assumes the existence of this dummy value.
c           Note: this array is overwritten
c
c    LDQG   INTEGER
c           Leading dimension of the array QG, as declared in the calling
c           program. Require  LDQG .GE. 2
c
c    COLQG  INTEGER
c           Number of columns of array QG.
c           A sufficiently large value of COLQG is  g  calculated as follows
c              let: mu = NN(1) = rank(B)
c                   a,b be non-negative integers such that  N = a*mu + b,  b<mu
c                   bsum = b*(b+1)/2,
c                   musum1 = (mu-1)*mu/2
c              then
c                       | ( a/2 )*( 1 + musum1 ) + mu - b ,       a even
c                   g = |
c                       | ( (a-1)/2 )*( 1 + musum1 ) + bsum ,     a odd
c
c    LDQH   INTEGER
c           Leading dimension of the array QH, as declared in the calling
c           program.
c           Require LDQH .GE. 3
c
c    COLQH  INTEGER
c           Number of columns of array QH.
c           A sufficiently large value of COLQH is  h  calculated as follows
c              let: mu, a, b, bsum1  be defined as above (see COLQG)
c                   musum = mu*(mu+1)/2
c              then
c                     | ( a/2 )*( musum*(a-2)/2 + bsum + 1 ) + mu,       a even
c                 h = |
c                     | ( (a-1)/2 )*( musum*(a-1)/2 + bsum + mu-b ) + b, a odd
c
c
c    Arguments out
c    -------------
c
c    NCMPLX INTEGER
c           Number of complex eigenvalues that were not allocated.
c
c    A      DOUBLE PRECISION array of DIMENSION (LDA,N)
c           The leading M by N part of this array contains the computed
c           gain matrix "F" 
c           If the given data has M>N, then the first M-N rows of F can be 
c           freely chosen and are neither computed nor stored, so that in
c           this case the leading N by N part of this array contains 
c           the last N rows of the computed gain matrix "F" 
c
c    B      DOUBLE PRECISION array of DIMENSION (LDA,M)
c           This array contains no useful information.
c
c    L      DOUBLE PRECISION array of DIMENSION (N)
c           Vector of allocated eigenvalues followed by eigenvalues 
c           that were not allocated.
c           Order of eigenvalues in L may differ from the original 
c           insofar as the eigenvalue origially stored as L(N) may
c           be moved to L(I), I .ne. N.   
c           Then the eigenvalues originally stored in L(I:N-1) will
c           be shifted to L(I+1:N), with no additional re-ordering.
c           The index I is returned to the calling program in INFO(2)
c
c    NN     INTEGER array of DIMENSION (kp1)
c           This array contains no useful information.
c
c    INFO   INTEGER array of DIMENSION (2)
c           INFO(1) returns number of successfully allocated eigenvalues.
c           INFO(2) returns index in L of eigenvalue originally stored
c                   as L(N)
c
c    QG     DOUBLE PRECISION array of DIMENSION (LDQG,COLQG)
c           Stores the Givens rotations used in the computation.
c
c    GCOL   INTEGER array of DIMENSION (COLQG)
c           Vector storing index associated with each stored rotation
c
c    QH     DOUBLE PRECISION array of DIMENSION (LDQH,COLQH)
c           Stores the Householder reflectors used in the computation.
c
c    HCOL   INTEGER array of DIMENSION (COLQH)
c           Vector storing index associated with each stored Householder
c
c    FCOL   INTEGER array of DIMENSION (N/2)
c           Vector of indeces indicating portions of feedback "F" to which
c           rotations comprising "P" have been applied. (see ref[1] for 
c           further details)
c
c
c    Work Space
c    ----------
c
c    IWORK  INTEGER array of DIMENSION (N)
c
c    RWORK  DOUBLE PRECISION array of DIMENSION (3*N)
c
c
c    Tolerances
c    ----------
c
c    TOL    DOUBLE PRECISION
c           Matrix elements with magnitudes less than TOL are considered zero.
c           If on entry TOL is less than the relative machine precision "eps",
c           it is reset to
c           TOL = (M+N)*||(B,A)||*eps
c                                   where ||.|| denotes the one-norm 
c           See LAPACK routine DLAMCH for details on computation of "eps"
c
c
c    Mode parameters
c    ---------------
c
c           None.
c
c
c    Warning Indicator
c    -----------------
c
c    IWARN  INTEGER
c           Unless M>N or the ranks of the staircase blocks do not sum to N
c           (see Warnings and Errors below), IWARN comtains 0 on exit
c           
c
c    Error Indicator
c    ---------------
c
c    IERR   INTEGER
c           Unless the routine detects an error (see next section),
c           IERR contains 0 on exit
c
c
c    Warnings and Errors detected by the Routine
c    ===========================================
c
c    IWARN = 1  On entry, M>N
c               In this case the first M-N rows of F can be freely chosen and 
c               will be neither computed nor stored.
c
c    IWARN = 2  Sum of ranks of staircase blocks is not equal to N.
c
c    IWARN = 3  On entry, conditions for  iwarn=1  and  iwarn=2 
c               both exist.
c
c
c    IERR = 1   Attempt to divide by zero or to solve singular system of 
c               equations. Here zero means any magnitude less than TOL
c
c    IERR = 2   Rank of the current deflated matrix is too low, indicating the
c               given system (B,A) is found to be too close (ie within TOL)
c               to an uncontrollable system.
c
c    IERR = 3   On entry, COLQG is too small for the number of Givens
c               transformations required for the computation.
c
c    IERR = 4   On entry, COLQH is too small for the number of Householder
c               transformations required for the computation.
c
c
c    Method
c    ======
c
c    An orthogonal matrix Q is computed along with the feedback matrix F so
c    that  Q'(A-BF)Q is in its real Schur form with specified eigenvalues.
c    The algorithm allocates two eigenvalues at a time in a series of double
c    steps. During the first double step, for example, the algorithm computes
c    orthogonal matrix Q1, say, and the first two columns of F*Q1, so that
c
c                  | a b | * * .. * |
c                  | c d | * * .. * |
c    Q1'(A-BF)Q1 = |-----|----------|
c                  |     | AA-BB*FF |
c
c    with  |a b|
c          |c d|  having two specified eigenvalues, and (BB, AA) being in
c    staircase form.  The orthogonal matrix Q is the product of N/2  or 
c    (N-1)/2 + 1 (depending on whether N is even or odd) orthogonal matrices
c    of the type Q1.
c
c
c    References
c    ==========
c
c    [1]  G. S. Miminis and C.C. Paige,
c         A double step algorithm for pole assignment of time invariant
c         multi-input linear systems using state feedback,
c         Technical Report 8908, Department of Computer Science,
c         Memorial University of Newfoundland, 1989.
c
c
c    Numerical Aspects
c    =================
c
c    The algorithm requires  O( N(N^2 + M(N-M)) )  operations  (see ref [1]).
c    The computation uses only real arithmetic, allocating complex eigenvalues
c    as conjugate pairs in double steps.
c
c
c    Contributors
c    ============
c 
c    G. Miminis, H. Roth  (Memorial University of Newfoundland, Canada)
c
c
c    Revisions
c    =========
c
c    1992 Oct 20
c
c  arguments
c     implicit         none
      integer          lda, n, ldb, m, ncmplx, nn(*)
      double precision A(lda,*), B(ldb,*), l(*), tol
      integer          ldqg, colqg, Gcol(*)
      integer          ldqh, colqh, Hcol(*), Fcol(*)
      double precision QG(ldqg,*), QH(ldqh,*), Rwork(*)
      integer          info(*), Iwork(*), iwarn, ierr
c
c
c  parameters
      double precision zero, one, two
      parameter (zero=0.0d0, one=1.0d0, two=2.0d0)
      integer ok
      parameter (ok=0)
c
c  local variables
      integer status, lnpos
      integer m0, m1, nm2
      integer i, j, j1, k, s, ss, itmp, infot, oddn, step, step1
      integer Bfree, Bfree1, free, free1, r, q, size, row, col
      integer Findex, Gindex, Hindex, Pindex, Pstart
      double precision atmp, cx, sx, lsum, lprod, f11, f12
      double precision eps
c      
c  intrinsic functions
      intrinsic min, max, abs
c
c  external functions
c   BLAS:     ddot
c   LAPACK:   dlamch
c   PACKAGE:  d1nrmU, d1nrmA
      double precision  ddot, dlamch, d1nrmU, d1nrmA
      external  ddot, dlamch, d1nrmU, d1nrmA
c
c  external subroutines
c   BLAS:
      external dcopy, drotg, drot, dswap
c   PACKAGE
      external dtinvb, dhhldr, dhhrfl, dabort
c     note:
c          dtinvb calls LAPACK routines dtrtrs, dtrcon
c
c
c  INITIALIZATION
c  **************
c
      status = 0
      lnpos = n
      iwarn = 0
      ierr  = 0
c
c  ====================================================================
c  Input arguments are checked by the driver subroutine.
c  Hence the following checks are commented out and not mentioned
c  in the documentation, but may be useful for further development.
c  
c  the arguments to dmevas are
c     n, m, ncmplx, A, lda,        B, ldb, l, nn, info,
c     QG, ldqg, colqg, Gcol, QH,   ldqh, colqh, Hcol, Fcol, tol
c     Iwork, Rwork, iwarn, ierr
c
c  check some input arguments
c     IF (colqh .LT. 1) ierr = -17
c     IF (ldqh .LT. 3) ierr = -16
c     IF (colqg .LT. 1) ierr = -13
c     IF (ldqg .LT. 2) ierr = -12
c     IF (ldb .LT. n) ierr = -7
c     IF (lda .LT. n) ierr = -5
c     IF( (ncmplx .LT. 0) .OR. ((ncmplx/2)*2 .NE. ncmplx)
c    &  .OR. (ncmplx .gt. n) ) ierr = -3
c     IF (m .LT. 1) ierr = -2
c     IF (n .LT. 1) ierr = -1
c
c     IF (ierr .lt. 0) THEN
c         GOTO 9000
c     ENDIF
c  ====================================================================
c
c  check that sum of ranks of staircase blocks is equal to N
c  if not set iwarn = 2
      k = 1
      itmp = N
c     do while ((nn(k) .ne. 0) .and. (k .le. N))
   70 IF ((nn(k) .ne. 0) .and. (k .le. N)) then
         itmp = itmp - nn(k)
         k = k + 1
         go to 70
      ENDIF
      IF (itmp .ne. 0) then
         iwarn = 2
      ENDIF
c
c     lnpos indicates position in final l-vector of initial l(n)
c     Bfree initialises to number of leading zero columns in B
c     free keeps number of leading zero rows in current A
c     step: A(step,step) is leading element in current A
c     Gindex is index to current Givens rotation in "Q"
c     Hindex is index to current Householder reflector in "Q"
c     Pindex is index to current Givens rotation in "P"
c     Findex points to first of columns of F which receive application
c          of current rotation in "P"
c     Pstart is index to first rotation (if any) in "P"
c     oddn = 1 if n is odd; 0 otherwise
c     m0 stores initial m and 
c     m1 is initialized to min(m,n). A warning is given if m>n. 
c
      m0 = m
      IF (m .LE. n) then
          m1 = m
      ELSE
          m1 = n
          iwarn = iwarn + 1
      ENDIF
      Bfree = m0-nn(1)
      Bfree1 = Bfree+1
      free = m1-nn(1)
      free1 = free+1
      step=1
      Gindex=0
      Hindex=0
      Findex=0
      Pstart=colqg+1
      Pindex=Pstart
      oddn = n-(n/2)*2
c
c  reset "tol" for numerical singularity, if necessary
c
c     .. calculate machine epsilon and store in "eps"
      eps = dlamch('e')
      IF (tol .lt. eps) then
c         .. compute 1-norm of system [B,A] and reset tol
          atmp = max(d1nrmU(B(1,Bfree1),ldb,nn(1)), d1nrmA(A,lda,n) )
          tol = (n+m1) * atmp * eps
      ENDIF
c
c
c  INITIAL IMMEDIATE ALLOCATIONS
c  *****************************
c
      s=nn(1)-nn(2)
      ss=(s/2)*2
      itmp = min(ss, ncmplx)
      IF (ss .GT. 0) then
            DO 110 i=2,itmp,2
               A(i-1,i-1) = A(i-1,i-1) - l(i-1)
               A(i,i-1) = A(i,i-1) + l(i)
               A(i,i) = A(i,i) - l(i-1)
               A(i-1,i) = A(i-1,i) - l(i)
  110       CONTINUE    
            ncmplx = ncmplx-itmp
            DO 120 i=itmp+1,ss
               A(i,i) = A(i,i) - l(i)
  120       CONTINUE
      ENDIF
c
c     If  s  and  n  both odd then allocate a real eigenvalue;
c     in particular allocate  l(n), the last eigenvalue in l,
c     shift l(s:n-1) to l(s+1:n), and set ss=s
c
      IF ((ss .NE. s) .AND. (oddn .NE. 0)) then
          atmp = l(n)
          call dcopy (n-s, l(s), -1, l(s+1), -1)
          l(s) = atmp
          A(s,s) = A(s,s) - atmp
          ss = s
          lnpos = s
      ENDIF
c
      IF (ss .GT. 0) then
          call dtinvb( B(step,Bfree1), ldb, nn(1), 
     &                 A, lda, ss, Iwork, Rwork, tol, infot )
          IF (infot .NE. ok) then
                   ierr = 1
                   call dabort(A, lda, m1, n, 0)
                   GOTO 9000
          ENDIF
c
c         relocate computed cols if free>0
          IF (free .GT. 0) then
              DO 140 j=1,ss
                  call dcopy(nn(1), A(1,j), -1, A(free1,j), -1)
                  call dcopy(free, zero, 0, A(1,j), 1)
  140         CONTINUE
          ENDIF
c
          IF (nn(1) .EQ. n) then
              status = n
              GOTO 9000
          ENDIF
      ENDIF
c
c     updates
      status = ss
      free = free+ss
      free1 = free+1
      Bfree = Bfree+ss
      Bfree1 = Bfree+1
      nn(1) = nn(1)-ss
      step = ss+1
c
c
c  DEFLATIONARY LOOP:
c  *****************
c
      nm2 = n-2
c     do while( step .LE. nm2  .AND.  nn(1)+nn(2) .GE. 2)
  200 IF ((step .LE. nm2) .AND. (nn(1)+nn(2) .GE. 2))  then
c               The second clause of the condition handles some 
c               pathological cases that can arise when SUM(nn(i)) < N
c
          s=nn(1)-nn(2)
c
          IF ( (s .EQ. 1) .AND. (oddn .EQ. 1) ) then
c
c            begin CASE 1.  Immediate single allocation of l(n)
c            ************
c
             atmp = l(n)
             call dcopy(n-step, l(step), -1, l(step+1), -1)
             l(step) = atmp
             lnpos = step
             A(step,step)=A(step,step)-atmp
c
             free1 = free+1
             Bfree1 = Bfree+1
             call dtinvb( B(step,Bfree1), ldb, nn(1),
     &             A(step,step), lda, 1, Iwork, Rwork, tol, infot )
             IF (infot .NE. ok) then
                 ierr = 1
                 call dabort(A, lda, m1, n, step-1)
                 GOTO 3000
             ENDIF
c
c            relocate 'F'
             IF (free1 .LT. step) then
                 call dcopy(nn(1), A(step,step), 1, A(free1,step), 1)
             ELSE
                 IF (free1 .GT. step) then
                    call dcopy(nn(1), A(step,step),-1, A(free1,step),-1)
                 ENDIF
             ENDIF
             call dcopy(free, zero, 0, A(1,step), 1)
c
c            updates
             status = status+1
             nn(1) = nn(2)
             step = step+1
             free = free1
             free1 = free+1
             Bfree = Bfree1
             Bfree1 = Bfree+1
             oddn = 0
c            end case 1: immediate single allocation
c
          ELSE IF (s .EQ. 2) then
c
c            begin CASE 2.  Immediate double allocation of l(step), l(step+1)
c            ************ 
c
             free1 = free+1
             Bfree1 = Bfree+1
             step1 = step+1
             IF (ncmplx .GT. 0) then
                A(step,step) = A(step,step)-l(step)
                A(step1,step) = A(step1,step)+l(step1)
                A(step1,step1) = A(step1,step1)-l(step)
                A(step,step1) = A(step,step1)-l(step+1)
             ELSE
                A(step,step) = A(step,step)-l(step)
                A(step1,step1) = A(step1,step1)-l(step1)
             ENDIF
             call dtinvb( B(step,Bfree1), ldb, nn(1),
     &             A(step,step), lda, 2, Iwork, Rwork, tol, infot )
             IF (infot .NE. ok) then
                 ierr = 1
                 call dabort(A, lda, m1, n, step-1)
                 GOTO 3000
             ENDIF
c
c            relocate 'F'
             IF (free1 .LT. step) then
                 call dcopy(nn(1), A(step,step),  1, A(free1,step),  1)
                 call dcopy(nn(1), A(step,step1), 1, A(free1,step1), 1)
             ELSE IF (free1 .GT. step) then
                 call dcopy(nn(1), A(step,step), -1, A(free1,step), -1)
                 call dcopy(nn(1), A(step,step1),-1, A(free1,step1),-1)
             ENDIF
                 call dcopy(free, zero, 0, A(1,step),  1)
                 call dcopy(free, zero, 0, A(1,step1), 1)

             IF (ncmplx .GT. 0) then
                 ncmplx = ncmplx-2
             ENDIF
c
c            update
             status = status+2
             nn(1) = nn(2)
             free = free+2
             free1 = free+1
             Bfree = Bfree+2
             Bfree1 = Bfree+1
             step = step+2
c
c            end CASE 2.  Immediate double allocation of l(step), l(step+1)
c
          ELSE
c
c            begin CASE 3: cases (s=0) OR (s=1 and n even)
c            ************
c
c            FIND NEXT r
             r=2
  300        CONTINUE
             IF ( nn(r) .EQ. nn(r+1) )  then
                r = r+1
                goto 300
             ENDIF
c
             IF ( r .GT. 2 ) then
c
c               begin CASE 3a.   regular double allocation  ( r>2 )
c               -------------
c
                q=nn(r)
                size = step-1 + nn(1) + (r-1)*q
                row = size-q+1
                col = row-q
c
c               Form row of N, taking advantage of upper Hessenberg structure,
c               and store in Rwork(1:n)
c               First non-zero in row-th row of A is in column col
c               First non-zero in row-th row of N is in column col-q
c               Use contiguous copy of row-th row of A
c
                IF (ncmplx .GT. 0) then
                    lsum = two * l(step)
                    lprod = l(step)**2 + l(step+1)**2
                ELSE 
                    lsum = l(step) + l(step+1)
                    lprod = l(step) * l(step+1)
                ENDIF
c
c               copy row-th row of A to Rwork(n+1:2n)
c               first non zero will be in Rwork(n+col)
                call dcopy(n, A(row,1), lda, Rwork(n+1), 1)
c
                ss = n+col
                itmp = col-q
                DO 350 j=itmp,row-1
                    k = j-itmp+1
c                     = # non-zeros in j-th column of A(col:size,col-q:size-q)
                    Rwork(j) = ddot(k, Rwork(ss), 1, A(col,j), 1) 
     &                           - lsum*A(row,j)
  350           CONTINUE
c
                k = 2*q
                Rwork(row) = lprod - lsum*A(row,j) +
     &                         ddot(k, Rwork(ss), 1, A(col,j), 1)
c
                DO 360 j=row+1,size
                    Rwork(j) = ddot(k, Rwork(ss), 1, A(col,j), 1)
     &                           - lsum*A(row,j)
  360           CONTINUE
c
c               P1:  first q-1 rotations
c
                IF (q .EQ. 1) then
c                   dummy rotation (to facilitate back transformation)
                    Gindex = Gindex+1
                    IF (Gindex .ge. Pindex) THEN
                        Gindex = Gindex-1
                        ierr = 3
                        call dabort(A, lda, m1, n, step-1)
                        GOTO 3000
                    ENDIF
                    Gcol(Gindex) = 0
                ELSE      
                    DO 420 i=size,size-q+2,-1
                        j = i-q
                        j1 = j+1
                        itmp = j-q
c
c                       compute rotation eliminating A(i,j) into A(i,j+1)
                        Gindex = Gindex+1
                        IF (Gindex .ge. Pindex) THEN
                            Gindex = Gindex-1
                            ierr = 3
                            call dabort(A, lda, m1, n, step-1)
                            GOTO 3000
                        ENDIF
                        call drotg( A(i,j1), A(i,j), cx, sx)
                        A(i,j) = zero
                        QG(1,Gindex) = cx 
                        QG(2,Gindex) = sx
                        Gcol(Gindex) = j
c
c                       post multiply A by computed rotation (to row i)
                        call drot( i-step, A(step,j1), 1,
     &                                  A(step,j), 1, cx, sx )
c
c                       pre multiply A by computed rotation (from column j-q)
                        call drot( n-itmp+1, A(j1,itmp), lda,
     &                                     A(j,itmp), lda, cx, sx )
c
c                       post multiply Rwork by computed rotation
                        call drot(1, Rwork(j1), 1, Rwork(j), 1, cx, sx)
  420               CONTINUE
                ENDIF
c
c               P2:  q-1 householders
c
                DO 480 i=row,row-q+2,-1
                    j=i-q-1
c                   compute Householder vector and store in QH(1:3,Hindex)
                    Hindex=Hindex+1
                    IF (Hindex .gt. colqh) THEN
                        Hindex = Hindex-1
                        ierr = 4
                        call dabort(A, lda, m1, n, step-1)
                        GOTO 3000
                    ENDIF
                    call dhhldr( A(i,j), lda, QH(1,Hindex))
                    Hcol(Hindex) = j
c
c                   post multiply A by computed Householder
                    DO 440 k=step,i-1
                        call dhhrfl( A(k,j), lda, QH(1,Hindex) )
  440               CONTINUE
c
c                   pre multiply A by computed Householder
                    DO 460 k=step,n
                        call dhhrfl( A(j,k), 1, QH(1,Hindex) )
  460               CONTINUE
c
c                   pre multiply B by computed Householder
                    DO 470 k=Bfree1,m0
                        call dhhrfl( B(j,k), 1, QH(1,Hindex) )
  470               CONTINUE
c
c                   post multiply Rwork by computed Householder
                    call dhhrfl( Rwork(j), 1, QH(1,Hindex) )
c
  480           CONTINUE
c
c               P3:   householder for row of N
c      
                j=col-q
                Hindex=Hindex+1
                IF (Hindex .gt. colqh) THEN
                    Hindex = Hindex-1
                    ierr = 4
                    call dabort(A, lda, m1, n, step-1)
                    GOTO 3000
                ENDIF
                call dhhldr (Rwork(j), 1, QH(1,Hindex))
                Hcol(Hindex) = j
c
c               post multiply A by computed Householder
                DO 500 i=step,col+1
                   call dhhrfl( A(i,j), lda, QH(1,Hindex))
  500           CONTINUE
c
c               pre multiply A by computed Householder
                DO 510 i=step,n
                   call dhhrfl( A(j,i), 1, QH(1,Hindex))
  510           CONTINUE
c
c               pre multiply B by computed Householder
                DO 520 k=Bfree1,m0
                    call dhhrfl( B(j,k), 1, QH(1,Hindex) )
  520           CONTINUE
c
c               Compute P4: product of  nn(1)-(r-4)*q  Householders
c
c               P4a: all but the last of P4's Householders:
c
                DO 650 i=row-q+1,step+q+3,-1
                     j=i-q-2
                     Hindex=Hindex+1
                     IF (Hindex .gt. colqh) THEN
                         Hindex = Hindex-1
                         ierr = 4
                         call dabort(A, lda, m1, n, step-1)
                         GOTO 3000
                     ENDIF
                     call dhhldr( A(i,j), lda, QH(1,Hindex) )
                     Hcol(Hindex) = j
c
c                    post multiply A by computed Householder
                     DO 600 k=step,i-1
                          call dhhrfl( A(k,j), lda, QH(1,Hindex) )
  600                CONTINUE
c
c                    pre multiply A by computed Householder
                     DO 610 k=step,n
                          call dhhrfl( A(j,k), 1, QH(1,Hindex) )
  610                CONTINUE
c
c                    pre multiply B by computed Householder 
                     DO 620 k=Bfree1,m0
                          call dhhrfl( B(j,k), 1, QH(1,Hindex) )
  620                CONTINUE
  650           CONTINUE
c
c               P4b: last householder if needed
c
                IF ( nn(1)+(r-4)*q .NE. 0 ) then
c
                    Hindex = Hindex+1
                    IF (Hindex .gt. colqh) THEN
                        Hindex = Hindex-1
                        ierr = 4
                        call dabort(A, lda, m1, n, step-1)
                        GOTO 3000
                    ENDIF
                    Hcol(Hindex) = step
c
                    IF ( nn(1) .EQ. q ) then 
c                       case nn(1) = q
                        call dhhldr(A(step+q+2,step),lda,QH(1,Hindex))
c                       post multiply A by computed Householder
                        DO 670 k=step,step+q+1
                            call dhhrfl( A(k,step), lda, QH(1,Hindex) )
  670                   CONTINUE
                    ELSE
c                       case nn(1) = q+1
                        QH(1,Hindex) = -one
                        QH(2,Hindex) = zero
                        QH(3,Hindex) = one

c                       post multiply A by permuting Householder
                        call dswap(n-step+1, A(step,step), 1,
     &                                       A(step,step+2), 1)
                    ENDIF
c
c                   pre multiply A by computed Householder
                    DO 690 k=step,n
                        call dhhrfl( A(step,k), 1, QH(1,Hindex) )
  690               CONTINUE
c
c                   pre multiply B by  computed Householder 
                    DO 700 k=Bfree1,m0
                        call dhhrfl( B(step,k), 1, QH(1,Hindex) )
  700               CONTINUE
                ENDIF
c
                step1 = step+1
                itmp = step+2
                call dtinvb( B(itmp,Bfree1),ldb,nn(1), 
     &                A(itmp,step),lda,2, Iwork, Rwork, tol, infot)
c
                IF (infot .NE. ok) then
                    ierr = 1
                    call dabort(A, lda, m1, n, step-1)
                    GOTO 3000
                ENDIF
c
c               relocate computed cols of F to first m rows of A
                IF (free .LT. step1) then
c                   { free1 < itmp }
                    call dcopy(nn(1), A(itmp,step), 1, 
     &                                  A(free1,step), 1)
                    call dcopy(nn(1), A(itmp,step1), 1, 
     &                                  A(free1,step1), 1)
                ELSE IF (free .GT. step1) then
                    call dcopy(nn(1), A(itmp,step), -1, 
     &                                  A(free1,step), -1)
                    call dcopy(nn(1), A(itmp,step1), -1, 
     &                                  A(free1,step1), -1)
                ENDIF
                call dcopy(free, zero, 0, A(1,step),  1)
                call dcopy(free, zero, 0, A(1,step1), 1)
c
c               end CASE 3a.   regular double allocation  ( r>2 )
c
             ELSE
c
c               begin CASE 3b.   Case r=2
c               -------------
c
                q = nn(2)
c
                IF ( s .EQ. 0 ) then
c
c                 begin 3b SUB-CASE  r=2 with nn(1) = nn(2)
c                 -----------------------------------------
c
c                 P1:  q-1 rotations
c      
                  DO 1020 j=step-1+q,step+1,-1
                     i=j+q
c
c                    compute rotation eliminating A(i,j) into A(i,j+1)
                     call drotg( A(i,j+1), A(i,j), cx, sx)
                     A(i,j)=zero
                     Gindex = Gindex+1
                     IF (Gindex .ge. Pindex) THEN
                         Gindex = Gindex-1
                         ierr = 3
                         call dabort(A, lda, m1, n, step-1)
                         GOTO 3000
                     ENDIF
                     QG(1,Gindex) = cx
                     QG(2,Gindex) = sx
                     Gcol(Gindex) = j
c
c                    post multiply A by computed rotation
                     call drot(i-step,A(step,j+1),1,
     &                             A(step,j),1, cx, sx )
c
c                    pre multiply A by computed rotation
                     call drot(n-step+1,A(j+1,step),lda,
     &                             A(j,step),lda, cx, sx )
c
c                    premultiply B by computed rotation
                     call drot(m0-Bfree,B(j+1,Bfree1),ldb,
     &                             B(j,Bfree+1),ldb, cx, sx )
 1020             CONTINUE
c
c
c                 P2: dummy Householder
                  Hindex = Hindex+1
                  IF (Hindex .gt. colqh) THEN
                      Hindex = Hindex-1
                      ierr = 4
                      call dabort(A, lda, m1, n, step-1)
                      GOTO 3000
                  ENDIF
                  Hcol(Hindex) = 0
c
c                 compute F10
                  IF( Bfree+2 .LE. m0 ) then
                      call dtinvb( B(step+2,Bfree+2),ldb,q-1, 
     &                             A(step+2,step),lda,2, 
     &                             Iwork, Rwork, tol, infot)
                      IF(infot .NE. ok) then
                         ierr = 1
                         call dabort(A, lda, m1, n, step-1)
                         GOTO 3000
                      ENDIF
c
c                     compute A10 - B01*F10
                      i=step
                      j=Bfree+2
                      A(i,i) = A(i,i) - ddot(q-1, B(i,j),ldb,
     &                                              A(i+2,i),1 )
                      A(i+1,i) = A(i+1,i) - ddot(q-1, B(i+1,j),ldb,
     &                                              A(i+2,i),1 )
                      A(i+1,i+1) = A(i+1,i+1) - ddot(q-1, B(i+1,j),ldb,
     &                                              A(i+2,i+1),1 )
                      A(i,i+1) = A(i,i+1) - ddot(q-1, B(i,j),ldb,
     &                                              A(i+2,i+1),1 )
                  ENDIF
c
                  step1 = step+1
                  IF ( (abs(B(step,Bfree1)) .LT. tol) .OR.
     &                 (abs(A(step1,step)) .LT. tol)      ) then
                          ierr = 1
                          call dabort(A, lda, m1, n, step-1)
                          GOTO 3000
                  ENDIF
c
c                 compute f1'
                  IF ( ncmplx .GT. 0 ) then
                      lsum = l(step) + l(step)
                      lprod= l(step)*l(step) + l(step1)*l(step1)
                  ELSE
                      lsum = l(step) + l(step1)
                      lprod= l(step)*l(step1)
                  ENDIF
c
                  atmp = A(step1,step1)
c
                  f11 = ( A(step,step)+atmp-lsum ) / B(step,Bfree1)
c
                  f12 = ( (atmp*(atmp-lsum)+lprod)/A(step1,step)
     &                  + A(step,step1) ) / B(step,Bfree1)
c
c                 relocate computed columns of F into A(1:m,:)
                  IF (free .LT. step) then
c                     free+2 < step+2
                      call dcopy(q-1, A(step+2,step), 1, 
     &                                   A(free+2,step), 1)
                      call dcopy(q-1, A(step+2,step1), 1, 
     &                                   A(free+2,step1), 1)
                  ELSE IF (free+2 .GT. step+2) then
                      call dcopy(q-1, A(step+2,step), -1, 
     &                                   A(free+2,step), -1)
                      call dcopy(q-1, A(step+2,step1), -1, 
     &                                   A(free+2,step1), -1)
                  ENDIF
                  A(free1,step)=f11
                  A(free1,step1)=f12
                  call dcopy(free, zero, 0, A(1,step), 1)
                  call dcopy(free, zero, 0, A(1,step1), 1)
c
c                 end CASE r=2 with nn(1) = nn(2)
c
                ELSE
c                 begin 3b SUB CASE  r=2 with nn(1) = nn(2)+1   (s=1)
c                 -------------------------------------------
c
c                 P1: q-1 rotations   { q = nn(2) }
c
                  IF (q .LE. 1) then
c                     dummy rotation if q-1 .LE. 0
                      Gindex = Gindex+1
                      IF (Gindex .ge. Pindex) THEN
                          Gindex = Gindex-1
                          ierr = 3
                          call dabort(A, lda, m1, n, step-1)
                          GOTO 3000
                      ENDIF
                      Gcol(Gindex) = 0
                  ELSE
                      DO 1120 j=step+q,step+2,-1
                          i=j+q
c
c                         compute rotation eliminating A(i,j) into A(i,j+1)
                          call drotg( A(i,j+1), A(i,j), cx, sx)
                          A(i,j)=zero
                          Gindex = Gindex+1
                          IF (Gindex .ge. Pindex) THEN
                              Gindex = Gindex-1
                              ierr = 3
                              call dabort(A, lda, m1, n, step-1)
                              GOTO 3000
                          ENDIF
                          QG(1,Gindex) = cx
                          QG(2,Gindex) = sx
                          Gcol(Gindex) = j
c
c                         post multiply A by computed rotation
                          call drot(i-step,A(step,j+1),1,
     &                             A(step,j),1, cx, sx )
c
c                         pre multiply A by computed rotation
                          call drot(n-step+1,A(j+1,step),lda,
     &                                       A(j,step),lda, cx, sx )
c
c                         premultiply B by computed rotation
                          call drot(m0-Bfree,B(j+1,Bfree1),ldb,
     &                                       B(j,Bfree1),ldb, cx, sx )
 1120                 CONTINUE
                  ENDIF
c
c
c                 P2: Householder interchanging cols step, step+2
c
                  Hindex=Hindex+1
                  IF (Hindex .gt. colqh) THEN
                      Hindex = Hindex-1
                      ierr = 4
                      call dabort(A, lda, m1, n, step-1)
                      GOTO 3000
                  ENDIF
                  QH(1,Hindex) = -one
                  QH(2,Hindex) = zero
                  QH(3,Hindex) = one
                  Hcol(Hindex) = step
c
c                 post multiply A by permuting Householder
                  call dswap(n-step+1,A(step,step),1,A(step,step+2),1)
c
c                 pre multiply A by permuting Householder
                  call dswap(n-step+1, A(step,step),lda, 
     &                                   A(step+2,step),lda)
c
c                 pre multiply B by permuting Householder 
                  call dswap(m0-Bfree, B(step,Bfree1),ldb,
     &                                   B(step+2,Bfree1),ldb)
c
c                 compute P: rotation to eliminate B(3,1) into B(3,2)
                  i=step+2
                  j=Bfree1
                  Pindex = Pindex-1
                  IF (Pindex .le. Gindex) THEN
                      Pindex = Pindex+1
                      ierr = 3
                      call dabort(A, lda, m1, n, step-1)
                      GOTO 3000
                  ENDIF
                  call drotg( B(i,j+1), B(i,j), cx, sx )
                  B(i,j) = zero
                  QG(1,Pindex) = cx
                  QG(2,Pindex) = sx
                  Gcol(Pindex) = j
                  Findex = Findex+1
                  Fcol(Findex) = step
c
c                 post multiply B by P
                  call drot( 2, B(step,j+1),1, B(step,j),1, cx, sx )
c
c                 compute F10
                  call dtinvb( B(step+2,Bfree+2),ldb,q, 
     &                  A(step+2,step),lda,2, Iwork, Rwork, tol, infot)
c
                  IF (infot .NE. ok) then
                      ierr = 1
                      call dabort(A, lda, m1, n, step-1)
                      GOTO 3000
                  ENDIF
c
c                 compute A10 - B01*F10
                  i=step
                  j=Bfree+2
                  A(i,i) = A(i,i) - ddot(q, B(i,j),ldb, A(i+2,i),1)
                  A(i+1,i) = A(i+1,i) - ddot(q, B(i+1,j),ldb,
     &                                               A(i+2,i),1   )
                  A(i+1,i+1) = A(i+1,i+1) - ddot(q, B(i+1,j),ldb,
     &                                               A(i+2,i+1),1 )
                  A(i,i+1) = A(i,i+1) - ddot(q, B(i,j),ldb,
     &                                               A(i+2,i+1),1 )
c
                  step1 = step+1
                  IF ( (abs(B(step1,Bfree1)) .LT. tol) .OR.
     &                 (abs(A(step,step1)) .LT. tol)      ) then
                          ierr = 1
                          call dabort(A, lda, m1, n, step-1)
                          GOTO 3000
                  ENDIF
c
c                 compute f1'
                  IF ( ncmplx .GT. 0 ) then
                      lsum = l(step) + l(step)
                      lprod= l(step)*l(step) + l(step1)*l(step1)
                  ELSE
                      lsum = l(step) + l(step1)
                      lprod= l(step)*l(step1)
                  ENDIF
c
                  atmp = A(step,step)
c
                  f11 = ( (atmp*(atmp-lsum)+lprod)/A(step,step1) +
     &                    A(step1,step) ) / B(step1,Bfree1)
c                  
                  f12 = ( atmp+A(step1,step1)-lsum ) / B(step1,Bfree1)
c
c                 relocate computed columns of F into A(1:m,:)
                  IF (free .LT. step) then
c                     { free+2 < step+2 }
                      call dcopy(q, A(step+2,step), 1, 
     &                                   A(free+2,step), 1)
                      call dcopy(q, A(step+2,step1), 1, 
     &                                   A(free+2,step1), 1)
                  ELSE IF (free .GT. step) then
                      call dcopy(q, A(step+2,step), -1, 
     &                                   A(free+2,step), -1)
                      call dcopy(q, A(step+2,step1), -1, 
     &                                   A(free+2,step1), -1)
                  ENDIF
                  A(free1,step)=f11
                  A(free1,step1)=f12
                  call dcopy(free, zero, 0, A(1,step), 1)
                  call dcopy(free, zero, 0, A(1,step1), 1)
c
c                 end CASE r=2 with nn(1) = nn(2)+1   (s=1)
c
                ENDIF
c               (3b r=2 subcase s=0, or subcase s=1)
c
c               update for case r=2
                free = free+1
                free1 = free+1
                Bfree = Bfree+1
                Bfree1 = Bfree+1
c
             ENDIF
c            (case 3a or 3b)
c
c            updates for case 3
             status = status+2
             IF (ncmplx .GT. 0) ncmplx=ncmplx-2
             Gcol(Gindex) = -Gcol(Gindex)
             Hcol(Hindex) = -Hcol(Hindex)
             nn(r-1) = nn(r-1) - 1
             nn(r) = nn(r) - 1
             step = step+2
c
c            end CASE 3: cases (s=0) OR (s=1 and n even)
c
          ENDIF
c         (case 1 or 2 or 3)
c
          goto 200
      ENDIF
c     end do !while (step .LE. nm2)  (end deflationary loop)
c
c  FINAL ALLOCATIONS
c  *****************
c
      step1 = step+1
      free1 = free+1
      Bfree1 = Bfree+1
c
      IF (nn(2) .EQ. 0) then
c         case resulting k=1   # of inputs = # of states
c
          IF ( nn(1) .EQ. 2 ) then
c             nn = [2,0,...,0]  Two eigenvalues remaining
c
             IF (ncmplx .GT. 0) then
                A(step,step) = A(step,step)-l(step)
                A(step1,step) = A(step1,step)+l(step1)
                A(step1,step1) = A(step1,step1)-l(step)
                A(step,step1) = A(step,step1)-l(step1)
             ELSE
                A(step,step) = A(step,step)-l(step)
                A(step1,step1) = A(step1,step1)-l(step1)
             ENDIF
c
             call dtinvb( B(step,Bfree1), ldb, nn(1),
     &             A(step,step), lda, 2, Iwork, Rwork, tol, infot )
c
             IF (infot .NE. ok) then
                 ierr = 1
                 call dabort(A, lda, m1, n, step-1)
                 GOTO 3000
             ENDIF
c
c            relocate computed columns of F into A(1:m,:)
             IF (free1 .LT. step) then
                 call dcopy(nn(1), A(step,step), 1, A(free1,step), 1)
                 call dcopy(nn(1), A(step,step1), 1, A(free1,step1), 1)
             ELSE IF (free1 .GT. step) then
                 call dcopy(nn(1), A(step,step),-1, A(free1,step),-1)
                 call dcopy(nn(1), A(step,step1),-1, A(free1,step1),-1)
             ENDIF
                 call dcopy(free, zero, 0, A(1,step), 1)
                 call dcopy(free, zero, 0, A(1,step1), 1)
c
             status = status+2
             IF (ncmplx .GT. 0) then
                 ncmplx = ncmplx-2
             ENDIF
c
          ELSE IF ( nn(1) .eq. 1  .and.  ncmplx .eq. 0 ) then
c            nn = [1,0,...,0]
c            procede to allocate one real eigenvalue
c
             A(step,step)=A(step,step)-l(step)
             call dtinvb( B(step,Bfree+1), ldb, nn(1), 
     &             A(step,step), lda, 1, Iwork, Rwork, tol, infot )
             IF (infot .NE. ok) then
                 ierr = 1
                 call dabort(A, lda, m1, n, step-1)
                 GOTO 3000
             ENDIF
c
c            relocate computed column of F into A(1:m,:)
             call dcopy(free, zero, 0, A(1,step), 1)
c            A(m,n) = A(n,n)
             A(m1,step) = A(step,step)
c
             status = status+1
c
c         otherwise
c            nn = [1,0,...,0] with 2 complex eigenvalues to be allocated, or
c            nn = [0,0,...,0]  
c            In either case the system is uncontrollable and no more 
c            allocations are possible. (recall we use only real arithmetic)
c            this is taken care of later
c
          ENDIF
c         (end case k=1)
c
      ELSE
c
c         case resulting k=2   one input, two states
c         nn = [1,1,0,...,0]
c
          step1 = step+1
          IF ( ncmplx .GT. 0 ) then
              lsum = l(step) + l(step)
              lprod= l(step)*l(step) + l(step1)*l(step1)
          ELSE
              lsum = l(step) + l(step1)
              lprod= l(step)*l(step1)
          ENDIF
c
          IF ( (abs(B(step,Bfree+1)) .LT. tol) .OR.
     &         (abs(A(step1,step)) .LT. tol)      ) then
                 ierr = 1
                 call dabort(A, lda, m1, n, step-1)
                 GOTO 3000
          ENDIF
c
          atmp = A(step1,step1)
c
          f11 = ( A(step,step)+atmp-lsum ) / B(step,Bfree+1)
c
          f12 = ( (atmp*(atmp-lsum)+lprod)/A(step1,step)
     &                  + A(step,step1) ) / B(step,Bfree+1)
c
c         relocate last two columns of F
          call dcopy( free, zero, 0, A(1,step), 1)
          call dcopy( free, zero, 0, A(1,step1), 1)
          A(m1,n-1)=f11
          A(m1,n) = f12
c
          status = status+2
          IF (ncmplx .GT. 0) then
             ncmplx = ncmplx - 2
          ENDIF
c
c         (end case k=2)       
      ENDIF
c     (end final allocations)
c
c     check if there were more eigenvalues in the given data
c     in that case system is uncontrollable
c     (some possibilities are
c        nn = [1,0,...,0] with 2 complex eigenvalues to be allocated, or
c        nn = [0,0,...,0]  
c        In either case the system is uncontrollable and no more 
c        allocations are possible. (recall we use only real arithmetic)
      IF ( status .ne. n ) THEN
         ierr = 2
         call dabort(A, lda, m1, n, step)
         GOTO 3000
      ENDIF
c
c
c  BACK TRANSFORMATION
c  *******************
c
 3000 CONTINUE
c
c     Apply P
c     -------
c     do while (Pindex .LT. Pstart) 
 3050 IF (Pindex .LT. Pstart) then
c         Apply transpose of rotation stored in QG(1,Pindex) to
c         rows Gcol(Pindex), Gcol(Pindex)+1  from cols Fcol(Findex) to n
          i = Gcol(Pindex)
          j = Fcol(Findex)
          cx = QG(1,Pindex)
          sx = QG(2,Pindex)
          call drot(n, A(i+1,j), lda, A(i,j), lda, cx, -sx)
          Findex = Findex-1
          Pindex = Pindex+1
          GOTO 3050
      ENDIF
c     end do !while (Pindex .LT. Pstart) 
c
c  Apply Q-inverse
c  ---------------
c     do while (Gindex .NE. 0)
 3100 IF (Gindex .NE. 0) then
c 
c        1. Apply inverse Householders for one step
c
         IF ( Hcol(Hindex) .EQ. 0 ) then
c            ignore dummy Householder
             Hindex = Hindex-1
         ELSE
c            apply (inverse of) Householders for one step
             IF (Hcol(Hindex) .LT. 0)  Hcol(Hindex) = -Hcol(Hindex)
c            do while (Hcol(Hindex) .GT. 0)
 3200        IF ((Hindex .GT. 0) .AND.(Hcol(Hindex) .GT. 0)) then
                 DO 3250 k=1,m1
                    call dhhrfl( A(k,Hcol(Hindex)), lda, QH(1,Hindex))
 3250            CONTINUE
                 Hindex = Hindex-1
                 GOTO 3200
             ENDIF
c            end do !while Hcol(Hindex) .GT. 0)
         ENDIF
c
c        2. Apply inverse Rotations for one step
c
         IF ( Gcol(Gindex) .EQ. 0 ) then
c            ignore dummy rotation
             Gindex = Gindex-1
         ELSE
c            transform
             IF (Gcol(Gindex) .LT. 0)  Gcol(Gindex) = -Gcol(Gindex)
c            do while (Gindex .GT. 0) .AND. (Gcol(Gindex) .GT. 0)
 3400        IF ((Gindex .GT. 0) .AND. (Gcol(Gindex) .GT. 0)) then
                 j = Gcol(Gindex)
                 cx = QG(1,Gindex)
                 sx = QG(2,Gindex)
                 call drot(m1, A(1,j+1), 1, A(1,j), 1, cx, -sx)
                 Gindex = Gindex-1
                 GOTO 3400
             ENDIF
c            end do !while (Gcol(Gindex) .GT. 0)
         ENDIF
c
      GOTO 3100
      ENDIF
c     end do !while (Gindex .NE. 0)
c
 9000 CONTINUE
c
c  TERMINATION
c  ***********
      info(1) = status
      info(2) = lnpos
      return
c
c ** last line of subroutine dmevas **
      end
c==========================================================
c==========================================================
c
        subroutine dabort(A, lda, m, n, allocd)
c
c       Purpose
c       =======
c       To zero the sub-matrix of A contained in 
c       rows 1 to M, columns ALLOCD+1 to N
c       That is to set  A(I,J) = 0.0  whenever
c       I is in the set {1,..,M}  AND  J is in the set {ALLOCD+1,..,N}
c       
c       Argument List
c       =============
c
c       Arguments In
c       ------------
c
c       A      DOUBLE PRECISION array of DIMENSION(LDA,N)
c              The leading M by N part of this array must contain
c              the matrix with elements to be zeroed.
c              Note: this array is overwritten.   (See Purpose)
c
c       LDA    INTEGER
c              Leading dimension of the array A, as declared in the
c              calling program.
c
c       M      INTEGER
c              The last row of A that will have elements set to zero
c              M .LE. LDA
c
c       N      INTEGER
c              The last column of matrix A that will have elements
c              set to zero.
c
c       ALLOCD INTEGER
c              The last column of matrix A that will NOT be changed, ie
c              columns 1 to ALLOCD are left unchanged.
c
c       Arguments Out
c       -------------
c
c       A      DOUBLE PRECISION array of DIMENSION(LDA,N)
c              The matrix A with  A(I,J)=0.0  whenever
c              I is in {1,...,M}  and  J is in {ALLOCD+1,...,N}
c
c       Workspace
c       ---------
c       None.
c
c       Tolerances
c       ----------
c       None.
c
c       Mode Parameters
c       ---------------
c       None.
c
c       Warning Indicator
c       -----------------
c       None.
c
c       Error Indicator
c       ---------------
c       None.
c
c
c       Warnings and Errors detected by the routine
c       ===========================================
c       None.
c
c       Method
c       ======
c       Successive calls to BLAS routine DCOPY overwrite, column by column,
c       the designated column elements with zero-vectors of length M
c
c       References
c       ==========
c       C. Lawson, R. Hanson, D. Kincaid, F. Krogh, "Basic Linear Algebra
c       Subprograms for Fortran Usage", ACM Trans. Math. Soft., v5 (1979),
c       pages 308-323.
c
c       Contributors
c       ============
c       H. Roth  (Memorial University of Newfoundland, Canada)
c       
c       Revisions
c       =========
c       1992 Oct 20
c
c
c  arguments
c       implicit none
        integer lda, m, n, allocd
        double precision A(lda,*)
c
c  parameters
        double precision   zero
        parameter (zero=0.0d0)
c  external subroutines
        external DCOPY
c
c  local variables
        integer j
c
        DO 100 j=allocd+1,n
            call dcopy(m, zero, 0, A(1,j), 1)
  100   continue
c
        return
        end
c
c
c==========================================================
c==========================================================
c
      subroutine dhhldr(x,incx,v)
c
c       Purpose
c       =======
c       To compute 3-vector v, with v(3)=1.0, so that 
c       the Householder reflector H, where  H = I-2*v*v'/v'*v 
c       is such that for the given vector x, of dimension 3,
c       H*x = [0 0 -s]'
c       where  s = sign(x(3)) * norm2(x)
c       In addition
c           v'*v/2.0 is computed and returned in v(3)
c           v is contiguous in memory
c           x(i) is overwritten with zero, i=1,2
c           x(3) is overwritten with -s 
c
c       Argument List
c       =============
c
c       Arguments In
c       ------------
c       X      DOUBLE PRECISION array of DIMENSION (at least 3)
c              The 3 elements  X(1), X(1+INCX), X(1+2*INCX)
c              must contain the vector whose two leading elements
c              will be overwritten with zero when multiplied from
c              the left by the computed Householder.
c              Note: This array is overwritten.
c
c       INCX   The stride for elements of the vector X
c
c
c       Arguments Out
c       -------------

c       X      DOUBLE PRECISION array of DIMENSION (at least 3)
c              X(1) is overwritten with zero.
c              X(1+INCX) is overwritten with zero.
c              X(1+2*INCX) is overwritten with -s, where
c                          s = sign(x(3)) * norm2(x), where
c                          x = (X(1), X(1+INCX), X(1+2*INCX))'
c
c       V      DOUBLE PRECISION array of DIMENSION (at least 3)
c              v is computed so that v(3)=1.0 and, for the given vector x,
c              H*x = [0 0 -s]', where  H = I-2*v*v'/v'*v 
c              Instead of the known value 1.0, v'*v/2.0 is returned in V(3)
c
c       Workspace
c       ---------
c       None.
c
c       Tolerances
c       ----------
c       None.
c
c       Mode Parameters
c       ---------------
c       None.
c
c       Warning Indicator
c       -----------------
c       None.
c
c       Error Indicator
c       ---------------
c       None.
c
c
c       Warnings and Errors detected by the routine
c       ===========================================
c       None.
c
c       Method
c       ======
c       sigma = sign(x(3))*norm2(x)
c       beta = x(3) + sigma
c       v(1)=x(1)/beta,  beta .ne. zero
c       v(2)=x(2)/beta,  beta .ne. zero
c       v(3)=1.0
c
c       References
c       ==========
c       Golub, G.H. and Van Loan, C.F., Matrix Computations, 2-nd ed.,
c       Johns Hopkins University Press, Baltimore, 1989, pp. 195-199.
c
c       Contributors
c       ============
c       H. Roth  (Memorial University of Newfoundland, Canada)
c       
c       Revisions
c       =========
c       1992 Oct 20
c
c  arguments
c     implicit none
      integer  incx
      double precision x(*), v(*)
c
c  parameters
      double precision zero, one, two
      parameter (zero=0.0d0, one=1.0d0, two=2.0d0)
c
c  functions
      double precision dnrm2
      external dnrm2
      intrinsic sign
c
c  local variables
      integer inc1, inc2
      double precision nrmx, b, s
c
      nrmx = dnrm2(3,x,incx)
      IF (nrmx .EQ. zero) then
             v(1)=zero
             v(2)=zero
             goto 999
c 
      else IF (incx .EQ. 1) then
             s = sign(nrmx, x(3))
             b = x(3) + s
             v(1)=x(1)/b
             v(2)=x(2)/b
             v(3) = ( one - (x(3)-s)/b) / two
             x(1) = zero
             x(2) = zero
             x(3) = -s
c 
      else   
             inc1 = 1+incx
             inc2 = 1+2*incx
             s = sign(nrmx, x(inc2))
             b = x(inc2)+s
             v(1) = x(1)/b
             v(2) = x(inc1)/b
             v(3) = (one-(x(inc2)-s)/b)/two
             x(1) = zero
             x(inc1) = zero
             x(inc2) = -s
      endif
c 
  999 continue
      return
      end
c
c
c==========================================================
c==========================================================
c
      subroutine dhhrfl(x,incx,v)
c
c       Purpose
c       =======
c       Overwrite x with H*x where 
c       H is the Householder reflector (I-2vv'/v'v), where
c       v is computed by subroutine DHHLDR (with v(3) restored to 1.0)
c
c       Argument List
c       =============
c
c       Arguments In
c       ------------
c       X      DOUBLE PRECISION array of DIMENSION (at least 3)
c              The 3-element vector x to be overwritten by H*x must 
c              be contained in X(1), X(1+INCX), X(1+2*INCX).
c              Note: This array is overwritten.
c
c       INCX   The stride for elements of the vector X
c
c       V      DOUBLE PRECISION array of DIMENSION (3)
c              A 3-element vector computed by DHHLDR
c
c
c       Arguments Out
c       -------------
c       X      DOUBLE PRECISION array of DIMENSION (at least 3)
c              Let
c              x = (X(1), X(1+INCX), X(1+2*INCX))' and y=H*x, 
c              where
c              H is the Householder reflector (I-2vv'/v'v), where
c                v is computed by subroutine DHHLDR, so that
c                v(3) is assumed equal to 1.0 and stores v'v/2.
c              Then
c              X(1) is overwritten with y(1).
c              X(1+INCX) is overwritten with y(2).
c              X(1+2*INCX) is overwritten with y(3).
c
c       Workspace
c       ---------
c       None.
c
c       Tolerances
c       ----------
c       None.
c
c       Mode Parameters
c       ---------------
c       None.
c
c       Warning Indicator
c       -----------------
c       None.
c
c       Error Indicator
c       ---------------
c       None.
c
c
c       Warnings and Errors detected by the routine
c       ===========================================
c       None.
c
c       Method
c       ======
c       x = (I-2vv'/v'v)x 
c         = x-2vv'x/v'v
c         = x-v'x(2v/v'v)
c
c       References
c       ==========
c       Golub, G.H. and Van Loan, C.F., Matrix Computations, 2-nd ed.,
c       Johns Hopkins University Press, Baltimore, 1989, pp. 195-199.
c
c       Contributors
c       ============
c       H. Roth  (Memorial University of Newfoundland, Canada)
c       
c       Revisions
c       =========
c       1992 Oct 20
c
c  arguments
c     implicit none
      integer  incx
      double precision x(*), v(*)
c
c  parameters
      integer one
      parameter (one=1.0d0)
c
c  external functions
      double precision ddot
      external ddot
c
c  local variables
      double precision t,tmp
      integer inc1, inc2
c
      tmp = v(3)
      v(3) = one
      t = ddot(3,v,1,x,incx)/tmp
c
      IF (incx .EQ. 1) then
          x(1)=x(1)-t*v(1)
          x(2)=x(2)-t*v(2)
          x(3)=x(3)-t
      else
          inc1 = 1+incx
          inc2 = 1+2*incx
          x(1) = x(1)-t*v(1)
          x(inc1) = x(inc1)-t*v(2)
          x(inc2) = x(inc2)-t
      endif
c 
      v(3) = tmp
c
  999 continue
      return
      end
c
c
c==========================================================
c==========================================================
c
      subroutine dtinvb( T,ldt,n, B,ldb,p, Iwork, Rwork, tol,infot )
c
c     Purpose
c     =======
c     Overwrite B (NxP) with solution to TX=B where T (NxN) is
c     upper triangular.  If T is numerically singular then no
c     attempt is made to compute X.
c
c     Arguments in
c     ============
c
c     T      DOUBLE PRECISION array of DIMENSION (LDT,N)
c            The matrix T must occupy the leading N rows by N columns 
c            of the array T
c
c     LDT    INTEGER
c            row dimension of array T, as declared in the calling program
c
c     N      INTEGER        
c            row and column dimension of matrix T
c            row dimension of matrix B
c            N .LE. LDT
c
c     B      DOUBLE PRECISION array of DIMENSION (LDB,P)
c            the matrix B must occupy the leading N rows by P columns of 
c            the array T
c
c     LDB    INTEGER
c            row dimension of array B, as declared in the calling program
c
c     P      INTEGER
c            column dimension of matrix B
c
c     Arguments Out
c     -------------
c
c     B      DOUBLE PRECISION array of DIMENSION (LDB,P)
c            leading N rows by P columns contains solution X of TX=B,
c            if T is nonsingular. Otherwise, B is unchanged from B on entry.
c
c     IWORK  INTEGER array of DIMENSION (N)  
c            work space required for condition estimator
c
c     RWORK  DOUBLE PRECISION array of DIMENSION (3*N)
c            work space required for condition estimator
c
c     Tolerances
c     ----------
c     TOL    DOUBLE PRECISION
c            Matrix elements with magnitude < TOL are considered zero
c
c     Mode Parameters
c     ---------------
c     None.
c
c     Warning Indicator
c     -----------------
c     INFOT  INTEGER
c            Unless T is non-singular "to working precision", 
c            INFOT contains 0 on exit
c
c     Error Indicator
c     ---------------
c     None.
c
c
c     Warnings and Errors detected by the routine
c     ===========================================
c     INFOT > 0 :  T is non-singular to working precision.
c
c     Method
c     ======
c     The LApack routine DTRCON is used to obtain a condition estimate for T
c     If the system is estimated to be "sufficiently well conditioned", the
c     right hand side matrix is solved column by column via repeated calls
c     to the LApack routine DTRTRS
c
c     References
c     ==========
c     Coleman, T.F. and Van Loan, C.F., Handbook for Matrix Computations,
c     SIAM, Philadelphia, 1988, pp. 144-145.
c
c     Contributors
c     ============
c     H. Roth  (Memorial University of Newfoundland, Canada)
c       
c     Revisions
c     =========
c     1992 Oct 20
c
c  ARGUMENTS
c     implicit none
      integer ldt, n, ldb, p, Iwork(*), infot
      double precision T(ldt,*), B(ldb,*), Rwork(*), tol
c
c  EXTERNAL SUBPROGRAMS
      external dtrtrs, dtrcon
c
c  LOCAL VARIABLES
      double precision rcond
c
      call dtrcon( '1', 'U', 'N', n, T, ldt, rcond, Rwork, Iwork, infot)
c
      IF ( rcond .lt. tol ) then
           infot = 1
      ELSE
           infot = 0
           call DTRTRS( 'U', 'N', 'N', n, p, T, ldt, B, ldb, infot )
      ENDIF
      return
      end
c
c
c==========================================================
c==========================================================
c
        double precision function d1nrmU( U, ldu, n)
c
c       Purpose
c       =======
c       To compute 1-norm of order N upper-triangular matrix U
c
c       Argument List
c       =============
c
c       Arguments In
c       ------------
c       U      DOUBLE PRECISION array of DIMENSION (LDU,N)
c              The leading N by N part of this array must contain
c              the upper triangular matrix U.
c              Elements outside the upper triangular part of matrix U
c              are not referenced.
c
c       LDU    INTEGER
c              Leading dimension of the array U, as declared in the
c              calling program.
c
c       N      INTEGER
c              Order of the matrix U.
c
c       Arguments Out
c       -------------
c       None.
c       The FUNCTION returns the 1-norm of the matrix U.
c
c       Workspace
c       ---------
c       None.
c
c       Tolerances
c       ----------
c       None.
c
c       Mode Parameters
c       ---------------
c       None.
c
c       Warning Indicator
c       -----------------
c       None.
c
c       Error Indicator
c       ---------------
c       None.
c
c
c       Warnings and Errors detected by the routine
c       ===========================================
c       None.
c
c       Method
c       ======
c       The 1-norm of the upper triangular part of each column is 
c       computed using BLAS routine DASUM. The maximum of these
c       is returned as the matrix 1-norm.
c
c       References
c       ==========
c       1. Golub, G.H. and Van Loan, C.F., Matrix Computations, 2-nd ed.,
c          Johns Hopkins University Press, Baltimore, 1989, pp. 53-57.
c
c       2. C. Lawson, R. Hanson, D. Kincaid, F. Krogh, "Basic Linear Algebra
c          Subprograms for Fortran Usage", ACM Trans. Math. Soft., v5 (1979),
c          pages 308-323.
c
c       Contributors
c       ============
c       H. Roth  (Memorial University of Newfoundland, Canada)
c       
c       Revisions
c       =========
c       1992 Oct 20
c
c  arguments
c       implicit none
        integer ldu, n
        double precision U(ldu,*)
c
c    local variables
        integer j
        double precision tnrm
c
c    external functions blas dasum
        double precision dasum
        external dasum
c
c    intrinsic functions
        intrinsic max
c
        tnrm = U(1,1)
        DO 100 j=2,n
            tnrm = max( dasum(j,U(1,j),1), tnrm )
  100   CONTINUE
c
        d1nrmU = tnrm
c
        return
        end
c
c
c==========================================================
c==========================================================
c
        double precision function d1nrmA( A, lda, n )
c
c       Purpose
c       =======
c       To compute 1-norm of order N general matrix A
c
c       Argument List
c       =============
c
c       Arguments In
c       ------------
c       A      DOUBLE PRECISION array of DIMENSION (LDA,N)
c              The leading N by N part of this array must contain
c              the matrix A.
c
c       LDA    INTEGER
c              Leading dimension of the array A, as declared in the
c              calling program.
c
c       N      INTEGER
c              Order of the matrix A.
c
c       Arguments Out
c       -------------
c       None.
c       The FUNCTION returns the 1-norm of the matrix A.
c
c       Workspace
c       ---------
c       None.
c
c       Tolerances
c       ----------
c       None.
c
c       Mode Parameters
c       ---------------
c       None.
c
c       Warning Indicator
c       -----------------
c       None.
c
c       Error Indicator
c       ---------------
c       None.
c
c
c       Warnings and Errors detected by the routine
c       ===========================================
c       None.
c
c       Method
c       ======
c       The 1-norm of of each column of A is computed using 
c       BLAS routine DASUM. The maximum of these is returned
c       as the matrix 1-norm.
c
c       References
c       ==========
c       1. Golub, G.H. and Van Loan, C.F., Matrix Computations, 2-nd ed.,
c          Johns Hopkins University Press, Baltimore, 1989, pp. 53-57.
c
c       2. C. Lawson, R. Hanson, D. Kincaid, F. Krogh, "Basic Linear Algebra
c          Subprograms for Fortran Usage", ACM Trans. Math. Soft., v5 (1979),
c          pages 308-323.
c
c       Contributors
c       ============
c       H. Roth  (Memorial University of Newfoundland, Canada)
c       
c       Revisions
c       =========
c       1992 Oct 20
c
c    arguments
c       implicit none
        integer lda, n
        double precision A(lda,*)
c
c    local variables
        integer j
        double precision tnrm
c
c    external functions blas dasum
        double precision dasum
        external dasum
c
c    intrinsic functions
        intrinsic max
c
        tnrm = dasum( n, A(1,1), 1 )
        DO 100 j=2,n
            tnrm = max( dasum(n,A(1,j),1), tnrm )
  100   CONTINUE
c
        d1nrmA = tnrm
c
        return
        end
