c  == ==================================================================
c
      subroutine dmvsli(n, m, ncmplx, gmax, hmax,
     &                     A,lda, B,ldb, eigs, kstair,
     &                     info, iwork, rwork, tol, iwarn, ierr)
c
c  == ==================================================================
c
c    Purpose
c    =======
c
c    To compute a real matrix F so that the "closed-loop" matrix (A - B*F)
c    has a specified set of eigenvalues.
c
c    Here A and B are real matrices such that the system (B,A) is in "upper
c    staircase" (or "controllability") form, with
c    staircase blocks in upper triangular form, and
c    the set of specified eigenvalues is self conjugate.
c
c
c    Argument List
c    =============
c
c    Arguments In
c    ------------
c
c    N      INTEGER.
c           Row and column dimension of matrix A,
c           row dimension of matrix B,
c           length of vector of eigenvalues EIGS.
c           N .ge. 1
c
c    M      INTEGER.
c           Column dimension of matrix B.
c           M .ge. 1
c
c    NCMPLX INTEGER
c           Number of complex eigenvalues in EIGS. 
c           0 .le. NCMPLX .le. N, and NCMPLX even.
c
c    GMAX   INTEGER.
c           Maximum number of Givens rotations to be used in the
c           computation.  A sufficient value of  GMAX  may be computed as
c           follows (see also HMAX below):
c           let: q = ifix(N/M) and
c                r = N-q*M  so that  
c                N = q*M+r  where q, r are non-negative integers and r < M
c                rsum = r*(r+1)/2
c                Msum1 = M*(M-1)/2
c           then
c                     |     (q/2)*(1 + Msum1) + M-r,   q even
c              GMAX = |
c                     | ((q-1)/2)*(1 + Msum1) + rsum,  q odd.
c
c    HMAX   INTEGER.
c           Maximum number of Householder transformations to be used in 
c           the computation.  A sufficient value of  HMAX  may be computed
c           as follows:
c           let  q, r, rsum, Msum1 be defined as for GMAX above. In addition,
c           let: Msum = M*(M+1)/2 
c           then
c                     |     (q/2)*(Msum*(q-2)/2 + rsum + 1) + M,   q even
c              HMAX = |
c                     | ((q-1)/2)*(Msum*(q-1)/2 + rsum + M-r) + r, q odd.
c
c
c           The following code computes GMAX and HMAX for given N and M
c
c              INTEGER   N, M
c              INTEGER   Q, Q2, R, RSUM, MSUM, MSUM1, GMAX, HMAX
c              LOGICAL   EVEN
c              .. assume N and M are initialised and carry on ..
c              Q = IFIX(N/M)
c              Q2 = IFIX(Q/2)
c              R = N-Q*M
c              RSUM = R*(R+1)/2
c              MSUM = M*(M+1)/2
c              MSUM1 = M*(M-1)/2
c              EVEN = (Q2*2 .EQ. Q)
c              IF (EVEN) THEN
c                  GMAX = (Q2)*(1 + MSUM1) + M-R
c                  HMAX = (Q2)*(1 + RSUM + (Q2-1)*MSUM) + M
c              ELSE
c                  GMAX = (Q2)*(1 + MSUM1) + RSUM
c                  HMAX = (Q2)*(M-R + RSUM + Q2*MSUM) + R
c              END IF
c
c           and the following declarations define less stringent but simpler
c           values of GMAX and HMAX
c
c              INTEGER   N, M
c              PARAMETER (N =  , M =  )
c              INTEGER   Q, R, RSUM, MSUM, MSUM1
c              PARAMETER (Q = N/M,  R = N-Q*M,  RSUM = R*(R+1)/2)
c              PARAMETER (MSUM = M*(M+1)/2,  MSUM1 = M*(M-1)/2)
c              INTEGER   GMAX
c              PARAMETER (GMAX = (Q/2)*(1 + MSUM1) + M*(R+2)/2)
c              INTEGER   HMAX
c              PARAMETER (HMAX = (Q/2)*(MSUM*Q/2 + RSUM + M-R) + M)
c
c
c    A      DOUBLE PRECISION array of DIMENSION (LDA,N).
c           The leading N by N part of this array must contain the state
c           transition matrix A in controllability (upper staircase) form,
c           with staircase blocks in upper triangular form. (See SLICOT
c           routine AB01CD).
c           Note: this array is overwritten.
c
c    LDA    INTEGER.
c           Row dimension of array A, as declared in the calling program
c           LDA .ge. N
c
c    B      DOUBLE PRECISION array of DIMENSION (LDB,M).
c           The leading N by M part of this array must contain the input
c           matrix B in controllability (upper staircase) form.
c           Note: this array is overwritten.
c
c    LDB    INTEGER.
c           Row dimension of array B, as declared in the calling program
c           LDB .ge. N.
c
c    EIGS   DOUBLE PRECISION array of DIMENSION (N).
c           Vector of eigenvalues to be allocated.
c           The complex eigenvalues (there are NCMPLX of them) must occur as
c           conjugate pairs.  They are stored in EIGS(1:NCMPLX),  and the
c           real eigenvalues (there are N-NCMPLX of them) are stored in
c           EIGS(NCMPLX+1:N)
c           Since the real and imaginary parts of a complex number 
c           also determine its conjugate, only one real part and one
c           imaginary part are stored for each pair of conjugates. These
c           parts are stored in successive elements of EIGS, with the real
c           parts having odd indices.
c
c           EXAMPLE:
c              To store the four complex eigenvalues
c                     (0.1, 0.2), (0.1, -0.2), (0.3, -0.4), (0.3, 0.4)
c              and the two real eigenvalues
c                     0.5, 0.6
c              EIGS may be initialized to
c                     0.1, 0.2, 0.3, -0.4, 0.5, 0.6
c
c           Observe that for odd i < NCMPLX,  EIGS(i) and EIGS(i+1) are the
c           real and imaginary parts, respectively, of either member of a
c           pair of complex conjugate eigenvalues, as required.
c
c           Note: this array is overwritten. (That is, it may be rearranged).
c
c    KSTAIR INTEGER array of DIMENSION (1+kmax).
c           where kmax = controllability index of the system [B,A],  
c                          i.e. the number of stairs in the staircase form.
c           The leading kmax elements must contain the ranks of B and the
c           staircase blocks of A, so that     
c           KSTAIR (1) = rank of B, 
c           KSTAIR (k) = rank of (k,k-1) block element of A, for k=2:kmax,
c                        and 
c           KSTAIR (kmax+1) = 0 is required by the routine.
c           Note: this array is overwritten.
c
c
c    Arguments Out
c    -------------
c
c    NCMPLX INTEGER.
c           Number of complex eigenvalues that were not allocated.
c           Complex eigenvalues are always allocated as conjugate pairs, so 
c           NCMPLX will always be even.
c
c    A      DOUBLE PRECISION array of DIMENSION (LDA,N).
c           The leading M by N part of this array contains the computed
c           gain matrix "F".
c           If the given data has M>N, then the first M-N rows of F can be 
c           freely chosen and are neither computed nor stored, so that in
c           this case the leading N by N part of this array contains 
c           the last N rows of the computed gain matrix "F". 
c
c    B      DOUBLE PRECISION array of DIMENSION (LDB,M).
c           This array contains no useful information.
c
c    EIGS   DOUBLE PRECISION array of DIMENSION (N).
c           Vector of allocated eigenvalues followed by eigenvalues 
c           that were not allocated, if any.
c           Order of eigenvalues in EIGS may differ from the original 
c           insofar as the eigenvalue origially stored as EIGS(N) may
c           be moved to EIGS(I), with I .ne. N.   
c           Then the eigenvalues originally stored in EIGS(I:N-1) will
c           be shifted to EIGS(I+1:N), with no additional re-ordering.
c           This can occur only if N is odd (and hence EIGS(N) is real)
c           The index I is returned to the calling program in INFO(2). 
c           (See INFO below).
c
c    KSTAIR INTEGER array of DIMENSION (kmax+1).
c           This array contains no useful information.
c
c    INFO   INTEGER array of DIMENSION (2).
c           INFO(1) returns number of successfully allocated eigenvalues.
c           INFO(2) returns index in EIGS of eigenvalue originally stored
c                   as EIGS(N), ie on exit EIGS(INFO(2)) contains the value
c                   that was stored in EIGS(N) on entry. (See also EIGS above).
c
c
c    Work Space
c    ----------
c
c    RWORK  DOUBLE PRECISION array of DIMENSION (3N + 2*GMAX + 3*HMAX).
c
c    IWORK  INTEGER array of DIMENSION (N + N/2 + GMAX + HMAX).
c
c
c    Tolerances
c    ----------
c
c    TOL    DOUBLE PRECISION.
c           Matrix elements with magnitudes less than TOL are considered zero.
c           If on entry TOL is less than the relative machine precision "eps",
c           it is reset to
c           TOL = (M+N)*||(B,A)||*eps
c                                   where ||.|| denotes the one-norm.
c           See LAPACK routine DLAMCH for details re "eps".
c
c
c    Mode Parameters
c    ---------------
c
c           None.
c
c
c    Warning Indicator
c    -----------------
c
c    IWARN  INTEGER.
c           Unless M>N, or the ranks of the staircase blocks do not sum to N
c           (see Warnings and Errors below), IWARN comtains 0 on exit.
c           
c
c    Error Indicator
c    ---------------
c
c    IERR   INTEGER.
c           Unless the routine detects an error (see next section),
c           IERR contains 0 on exit.
c
c
c    Warnings and Errors detected by the Routine
c    ===========================================
c
c    IWARN = 1  On entry, M > N.
c               In this case the first N-M rows of F can be freely
c               chosen and will not be stored.
c
c    IWARN = 2  Sum of ranks of staircase blocks is not equal to N.
c
c    IWARN = 3  On entry, conditions for  iwarn=1  and  iwarn=2 
c               both exist.
c
c
c    IERR < 0   IERR = -j indicates a problem with the j-th argument
c               on entry.  Specifically:
c               IERR = -1   On entry,  N < 1
c               IERR = -2   On entry,  M < 1
c               IERR = -3   On entry,  NCMPLX < 0
c                                 or   NCMPLX > N
c                                 or   NCMPLX is an odd number
c               IERR = -4   On entry,  GMAX < 1
c               IERR = -5   On entry,  HMAX < 1
c               IERR = -7   On entry,  LDA < N
c               IERR = -9   On entry,  LDB < N
c
c    IERR = 1   Signifies attempt to divide by zero (ie a magnitude
c               less than TOL), or to solve a numerically singular
c               system of equations.
c
c    IERR = 2   During eigenvalue assignment a rank defficiency is
c               discovered in one of the staircase blocks, indicating
c               the system (B,A) is uncontrollable and assignment of
c               eigenvalues can proceed no farther.
c
c    IERR = 3   Signifies insufficient storage space for Givens rotations.
c               The quantity  GMAX  needs to be increased.
c
c    IERR = 4   Signifies insufficient storage space for Householder
c               transformations.  The quantity  HMAX  needs to be increased.
c
c
c    Method
c    ======
c
c    An orthogonal matrix Q is computed along with the feedback matrix F so
c    that  Q'(A-BF)Q is in its real Schur form with specified eigenvalues.
c    The algorithm allocates two eigenvalues at a time in a series of double
c    steps. During the first double step, for example, the algorithm computes
c    orthogonal matrix Q1, say, and the first two columns of F*Q1, so that
c
c                  | a b | * * .. * |
c                  | c d | * * .. * |
c    Q1'(A-BF)Q1 = |-----|----------|
c                  |     | AA-BB*FF |
c
c    with  |a b|
c          |c d|  having two specified eigenvalues, and (BB, AA) being in
c    staircase form.  The orthogonal matrix Q is the product of N/2  or 
c    (N-1)/2 + 1 (depending on whether N is even or odd) orthogonal matrices
c    of the type Q1.
c
c
c    References
c    ==========
c
c    [1]  G. S. Miminis and C.C. Paige,
c         A double step algorithm for pole assignment of time invariant
c         multi-input linear systems using state feedback,
c         Technical Report 8908, Department of Computer Science,
c         Memorial University of Newfoundland, 1989.
c
c
c    Numerical Aspects
c    =================
c     
c    The computation uses only real arithmetic, allocating complex eigenvalues
c    as conjugate pairs in "double steps".
c
c    The algorithm requires O( n(n**2 + m(n-m)) ) operations
c    (see ref [1]).
c
c
c    Contributors
c    ============
c
c    G. Miminis and H. Roth  (Memorial University of Newfoundland, Canada)
c
c
c    Revisions
c    =========
c
c    1992 Oct 19
c
c  == ==================================================================
c  == ==================================================================
c
c  declarations
c  ============
c
c     implicit none
c
c  arguments
      integer          n, m, ncmplx, gmax, hmax, lda, ldb
      double precision A(lda,*), B(ldb,*), eigs(*)
      integer          kstair(*), info(*), iwork(*)
      double precision rwork(*), tol
      integer          iwarn, ierr
c
c  parameters
      double precision  dzero
      parameter        (dzero = 0.0d0)
      integer           izero
      parameter        (izero = 0)
c
c  external subroutines
      external dcopy, dmevas
c
c  local variables
      integer i, ilen, rlen
c
c
c  code starts here
c  ================
c
c  initialize
c  ==========
        info(1) = 0
        info(2) = 0
        iwarn = 0
        ierr = 0
c
c  check some input arguments 
c  ==========================
c  set ierr = -k if we find a problem with the k-th argument
c  the arguments are
c     (n, m, ncmplx, gmax, hmax   A, lda, B, ldb, eigs, 
c         kstair, info, iwork, rwork, tol,    iwarn, ierr)
      IF( ldb .LT. n ) ierr = -9
      IF( lda .LT. n ) ierr = -7
      IF( hmax .LT. 1 ) ierr = -5
      IF( gmax .LT. 1 ) ierr = -4
      IF( (ncmplx .LT. 0) .OR. ((ncmplx/2)*2 .NE. ncmplx)
     &  .OR. (ncmplx .gt. n) ) ierr = -3
      IF( m .LT. 1 ) ierr = -2
      IF( n .LT. 1 ) ierr = -1
c
c  That's all we can check.  Quick return if we found a problem
      IF( ierr .lt. 0 ) GOTO 9000
c
c  clear the workspace
c  ===================
      rlen = 2*gmax + 3*hmax + 3*n
      ilen = gmax + hmax + n/2 + n
c
      call dcopy( rlen, dzero, 0, rwork, 1 )
c
      do 60 i = 1, ilen
          iwork(i) = izero
   60 continue
c
c  summarize the workspace partitioning
c  ====================================
c     .. QG    starts at rwork(1);                has dimension(2,gmax)
c     .. QH    starts at rwork(1+2*gmax);         has dimension(3,hmax)
c     .. Rwork starts at rwork(1+2*gmax+3*hmax);  has length 3N
c              min length(rwork) = 2*gmax+3*hmax+3*N
c     .. GCOL  starts at iwork(1);                has length gmax
c     .. HCOL  starts at iwork(1+gmax);           has length hmax
c     .. FCOL  starts at iwork(1+gmax+hmax);      has length N/2
c     .. Iwork starts at iwork(1+gmax+hmax+N/2);  has length N
c              min length(iwork) = gmax+hmax+N/2+N
c
c  recall the parameter sequence
c  =============================
c     subroutine dmevas(n,m,ncmplx, A,lda, B,ldb, l, nn, info,
c    &                QG,ldqg, colqg, Gcol,
c    &                QH,ldqh, Hcol, Fcol,
c    &                tol, Iwork, Rwork, iwarn, ierr)
c
c  do the job
c  ==========
      call dmevas (n,m,ncmplx, A,lda, B,ldb, eigs, kstair, info,
     &            rwork(1), 2, gmax, iwork(1),
     &            rwork(1+2*gmax), 3, hmax, iwork(1+gmax),
     &            iwork(1+gmax+hmax), tol, iwork(1+gmax+hmax+N/2),
     &            rwork(1+2*gmax+3*hmax), iwarn, ierr)
c
 9000 continue
      return
      end
