        SUBROUTINE SSEVAS(N,A,LDA,BINP,EIGVL,IEIGAL,IWRKDX,F,
     +                          IWORK,RWORK,TOL,IFORM,IWARN,IERR)

C     PURPOSE
C
C     Given an  nxn  matrix  A,  an  nx1  column vector  b,  and a
C     selfconjugate set  leig  of  n  numbers, the program 
C     calculates a  1xn  row vector  f^T  such that  (A - b * f^T)
C     has eigenvalues  leig.  A  and  b  may be general, or in 
C     "controllability" form, where  A  is upper hessenberg and  b
C     has only one non-zero element in  b(1).
C
C
C     ARGUMENTS IN
C
C       N - INTEGER.
C           The order of the matrix A and the number of elements in
C           arrays BINP, EIGVL, and F.
C           N .GE. 1
C
C       A - REAL array of DIMENSION (LDA,N).
C           The leading N by N part of this array contains the
C           matrix A.  If IFORM = .TRUE. then the matrix A is
C           assumed to be unreduced upper Hessenberg.
C           NOTE that this array is overwritten.
C
C     LDA - INTEGER.
C           The leading dimension of array A as declared in the calling
C           program.
C           LDA .GE. N
C
C    BINP - REAL array of DIMENSION (N).
C           This array contains the input vector B.
C           NOTE that this array is overwritten.
C
C
C   EIGVL - REAL array of DIMENSION (2,N).
C           This array contains the eigenvalues to be allocated.
C           EIGVL(1,k) holds the real part of the eigenvalue, while
C           EIGVL(2,k) holds the imaginary part.
C
C     ARGUMENTS OUT
C
C  IEIGAL - INTEGER.
C           Returns the number of eigenvalues actually allocated.
C
C  IWRKDX - INTEGER.
C           Returns the partitioning of the real work space.
C
C             - RWORK(IWRKDX): working storage for reflectors.
C
C             - RWORK(1) to RWORK(IWRKDX-1):  controllability form
C               workspace. (If IFORM = .FALSE.)
C
C
C       F - REAL array of DIMENSION(N).
C           Returns the solution vector F.
C
C
C     WORK SPACE
C
C   RWORK - REAL array of DIMENSION at least (kwrk1),
C           where kwrk1 >= (3*N**2 - N - 2)/2 if IFORM = .FALSE.,
C           and   kwrk1 >= N**2 - 2*N + 1  if IFORM = .TRUE..
C
C   IWORK - INTEGER array of DIMENSION at least (kwrk2),
C           where kwrk2 >= (N**2 - 2*N + 1)/4 
C
C
C     TOLERANCES
C
C     TOL - REAL.
C           The tolerance used in testing for controllability.  
C           If positive, the subroutine will use the given value 
C           as the tolerance.  If the given value of TOL is 0.0 
C           the subroutine will automatically calculate the 
C           tolerance.  (See note 3 below.)  The calculated 
C           tolerance is returned via this parameter.
C                                    
C           The tolerance used in testing for controllability is:
C          
C                    tol = n*||(b,A)||*u ,
C
C           where  u  is the machine unit and ||.|| is the 1-norm.
C
C
C
C     MODE PARAMETERS
C
C      IFORM - LOGICAL.
C              Specifies the input form as follows:
C
C              IFORM = .TRUE.,  Array A and vector B are in
C                      "controllability" form, where A is unreduced
C                       upper Hessenberg, and B is identically 0
C                       except for the fist element.
C
C              IFORM = .FALSE., Array A and vector B are general,
C                       with no specific form.
C
C     WARNING INDICATOR
C
C      IWARN - INTEGER.
C              Unless the routine detects a non-fatal error,
C              (see next section), IWARN contains 0 on exit.
C
C
C     ERROR INDICATOR
C
C       IERR - INTEGER.
C              Unless the routine detects a fatal error (see 
C              next section), IERR contains 0 on exit.
C
C     WARNINGS AND ERRORS DETECTED BY THE ROUTINE
C
C     IERR = -1 : On entry, N .LT. 1 or LDA .LT. N 
C
C     IERR =  1 : On entry, TOL .LT. 0.0
C
C     IERR =  0 : IWARN = 1
C                 An eigenvalue was found in array EIGVL that is
C                 not real or part of a complex conjugate pair.
C                 The number of eigenvalues actually allocated 
C                 is returned in IEIGAL, and the vector F is 
C                 returned such that (A - b*f) has those
C                 eigenvalues. (See notes (1) and (2) below.)
C
C     IERR =  0 : IWARN = 2
C                 Indicates the system is close to uncontrollable
C                 using the tolerance returned in TOL.
C                 The number of eigenvalues actually allocated 
C                 is returned in IEIGAL, and the vector F is 
C                 returned such that (A - b*f) has those
C                 eigenvalues. (See note (2) below.)
C
C
C     NOTES ON USAGE:
C
C       1) SSEVAS only takes eigenvalues that are real or in
C               complex conjugate pairs.  These are passed to the
C               subroutine via EIGVL, dimensioned as EIGVL(2,N).
C               EIGVL(1,x) holds the real part, while EIGVL(2,x) holds
C               the imaginary part.  Real valued eigenvalues can be in
C               any order.  Conjugate pairs of eigenvalues must be 
C               stored contiguously, with the first of the pair on an 
C               odd-numbered index in the array EIGVL. For example,
C               the eigenvalues 1,2,3, and 1 +/- 4 can be stored:
C
C                       [1.0,0.0]
C                       [2.0,0.0]
C                       [1.0,+4.0]
C                       [1.0,-4.0]
C                       [3.0,0.0]
C
C               but the following would generate an error:
C
C                       [1.0,0.0]
C                       [1.0,+4.0]
C                       [1.0,-4.0]
C                       [2.0,0.0]
C                       [3.0,0.0]
C
C               since the subroutine working with two eigenvalues
C               at a time, will work with eigenvalues  1.0  and
C               1.0 + i*4.0 , which is not allowed. An easy way
C               to list the eigenvalues could be to first list
C               all the complex eigenvalues, followed by the real
C               eigenvalues.
C
C       2) Before allocating a pair of eigenvalues, the subroutine
C               checks to see if the current matrix is controllable.
C               If this is not the case, no further allocations will
C               take place, and the subroutine will print a warning
C               indicating the number of successful allocations.
C               The resulting matrix (A-b*f) will have those 
C               eigenvalues allocated; the remainder will be undefined.
C
C       3) The SSEVAS routine uses the machine unit to calculate
C               a tolerance for determining controllability.  If
C               the input parameter TOL is identically 0.0, the
C               internal routine SGTTOL automatically calculates
C               the machine unit and uses it to compute a tolerance.
C               In computing the machine unit,we have adopted a
C               the code recommended by Gentleman and Marovich in
C               [CACM 17, p. 276].  The user however has the option
C               of supplying the tolerance.  The parameter TOL
C               returns the tolerance actually used in the
C               calculations.
C
C       4) This subroutine makes extensive use of the level-1 BLAS
C               (Basic Linear Algebra Subroutines) package.  The
C               BLAS-1 subroutines used are:
C
C               SASUM, SAXPY, SCOPY, SDOT, SNRM2, SSCAL, ISAMAX
C
C               These routines are available from various sources,
C               such as the IMSL subroutine library.  
C       
C
C
C     METHOD
C
C     In what follows,  Q^T  will denote the transpose of a matrix
C     Q.  The subroutine  _SEVAS  is an implementation of the 
C     algorithm in [1], and it is briefly described in the 
C     following steps:
C 
C 
C     (1) If (b,A) is in general form, transform to
C         controllability form using an orthogonal matrix  U:
C 
C                 (b,A) = U^T * (b,A) * ( 1 |   )
C                                       (-------) .
C                                       (   | U )
C 
C     (2) For each pair of eigenvalues (if they are complex they
C         appear as conjugates ),  l1=leig(i), l2=leig(i+1), 
C         with  i  going from  1  to  n-2  or  n-1,  with step  2,
C         perform the following:
C 
C             (2.1) Compute the last row of the matrix
C 
C                         (A - l1*I)*(A - l2*I)
C 
C                   and form a Householder transformation  P
C                   to eliminate the two nonzero off diagonal
C                   elements of this row into the diagonal element.
C 
C             (2.2) Compute  A = P^T * A * P, and  b = P^T * b.
C 
C             (2.3) Compute orthogonal matrix  Q  to transform A to
C                   the following form:
C 
C                                  (   H1   |    *    )
C                   Q^T * A * Q =  ( -----------------)
C                                  (    h   |         )
C                                  ( -------|    H2   )
C                                  (    0   |         )
C 
C                   where  H1  is a  2x2  matrix,  h  is a  2x1
C                   vector, and  H2  an  (n-i+1)x(n-i+1)  upper
C                   Hessenberg matrix.
C 
C                   Also transform  Q^T * b = ( x )
C                                             ( x )
C                                             ( g )
C 
C                   where  x  represents scalar, and  g  is a vector
C                   with  g(1)  being its only nonzero element.
C 
C             (2.4) Calculate the (i,i+1) elements of  f^T * P * Q
C                   such that  h = [0,0]. H1  now has eigenvalues
C                   l1, l2.
C 
C             (2.5) Repeat step 2 using  A = H2,  and  b = g.
C 
C     (3) If A is of size two  (respectively one), calculate the last
C         two (respectively  one) element(s) of the current  f^T
C         directly from  the current  (b,A)  and  leig(n-1,n)
C         (respectively  leig(n) ).
C 
C     (4) Transform the computed  f^T  using the history of  P  and  Q
C         transformations from step 2, as well as transformation  U
C         from step 1. The transformations should be applied in order
C         reverse of that they were computed with. The resulting  f
C         is the desired vector so that the eigenvalues of
C         (A - b * f^T)  are those in  leig.
C 
C 
C     REFERENCES
C 
C     [1] Miminis, S. G. & Paige C. C., "A double-step algorithm for
C         pole assignment of single-input linear systems", accepted
C         for publication by the SIAM journal on Matrix Analysis and
C         Applications.
C 
C 
C     NUMERICAL ASPECTS
C 
C     The algorithm requires  O(8/3 x n**3)  operations,  O(2 x n**2)
C     memory words, and is backward stable (see [1]).
C 
C 
C 
C     CONTRIBUTORS
C 
C 
C     G. Miminis and M. Reid (Memorial Univ. of Newfoundland, Canada).
C 
C 
C     REVISIONS
C
C     1991, October 14.
C 
C*********************************************************************

C
C.......scalar arguments 
C
        INTEGER N,LDA,IEIGAL,IWRKDX,IWARN,IERR
        REAL TOL
        LOGICAL IFORM
C
C.......array arguments
C
        INTEGER IWORK(*)
        REAL A(LDA,*),BINP(N),F(N),RWORK(*)
        REAL EIGVL(2,N)
C
C.......local scalars
C
        INTEGER ISIZE,LE,IREFC,IOLDRF,II,K
        REAL S1,S2,LM
C
C.......external functions
C
        REAL SGTTOL
        EXTERNAL SGTTOL
        LOGICAL SEIGCK 
        EXTERNAL SEIGCK
C
C.......external subroutines
C
        EXTERNAL SCNTMK, SAPPSH, SCRH, SPREM, SPOST
C
C.......intrinsic functions
C
        INTRINSIC ABS


C
C.......set initial values for error and warning flags
C

        IWARN = 0
        IERR = 0

C
C.......check error conditions
C
        IF(N .LT. 1  .OR.  LDA .LT. N) THEN
             IERR = -1
             RETURN
        ENDIF

C
C.......set index to partition work space
C
C            RWORK(1) to RWORK(IWRKDX-1):  controllability form
C            workspace (If IFORM = .FALSE.)
C
C            RWORK(IWRKDX): working storage for reflectors

        IF(IFORM) THEN
             IWRKDX = 1
        ELSE 
             IWRKDX = (N*N + 3*N - 4)/2 + 1
        ENDIF

C
C.......convert general form to controllability form
C

        IF(.NOT. IFORM) CALL SCNTMK(A,N,LDA,BINP,RWORK)


C
C.......calculate tolerance for controllability determination
C

        TOL = SGTTOL(A,N,LDA,BINP,TOL,IERR)
        IF(IERR .NE. 0) RETURN

C
C.......clear zero elements of system
C
        F(1) = 0.0
        F(2) = 0.0
        DO 50 I=3,N
            F(I) = 0.0
            DO 50 J=1,I-2
                A(I,J) = 0.0
50      CONTINUE
            


C
C.......start deflation: continue by double-step until problem is
C               smaller than size 3
C

        IEIGAL = 0
        IREFC = 0
        ISIZE = N
100     IF ( ISIZE .LE. 2) GOTO 200
        
        LE = N - ISIZE + 1
        IOLDRF = IREFC


C
C.......check eigenvalues for complex conjugates
C
        IF(.NOT. SEIGCK(EIGVL,LE,TOL)) THEN
                IWARN = 1
                GOTO 900
        ENDIF



C
C.......apply reflector containing eigenvalues to be allocated.
C

        CALL SAPPSH(A,LE,N,LDA,EIGVL,RWORK(IWRKDX),IWORK,IREFC)


C
C.......transform A to block upper hessenberg.
C

        CALL SCRH(A,LE,N,LDA,RWORK(IWRKDX),IWORK,IREFC)



C
C.......apply latest reflector to binp (vector of inputs)
C
        BINP(LE+1) = 0.0
        BINP(LE+2) = 0.0
        CALL SPREM(BINP,LDA,LE,1,1,RWORK(IWRKDX+(IREFC-1)*4),3)


C
C......check system for controllability
C
        IF(ABS(BINP(LE+2)) .LT. TOL) THEN
                IWARN = 2
                IREFC = IOLDRF
                F(LE) = 0.0
                F(LE+1) = 0.0
                GOTO 900

C
C.......specify le,le+1 elements of f
C
        ELSE
                F(LE) = A(LE+2,LE) / BINP(LE+2)
                F(LE+1) = A(LE+2,LE+1) / BINP(LE+2)
        ENDIF


C
C.......reduce problem size for next step
C
        ISIZE  = ISIZE - 2
        IEIGAL = IEIGAL + 2

        GOTO 100


200     CONTINUE

C
C.......allocate eigenvalues for matrices of size one or two.
C
        IF (ISIZE .EQ. 1) THEN
C
C...........check for controllability
C
            IF(ABS(BINP(N)) .LT. TOL) THEN
                IWARN = 2
                F(N) = 0.0
                GOTO 900
            ELSE
                F(N) = (A(N,N) - EIGVL(1,N))/BINP(N)
                IEIGAL = IEIGAL + 1
            ENDIF

        ELSE

C
C...........check for controllability  and conjugate 
C           eigenvalues (size 2)
C
            IF(.NOT. SEIGCK(EIGVL,LE,TOL) .OR.
     +                  ABS(BINP(N-1)) .LT. TOL .OR.
     +                  ABS(A(N,N-1)) .LT. TOL) THEN
                IF(ABS(BINP(N-1)) .LT. TOL .OR.
     +                  ABS(A(N,N-1)) .LT. TOL) THEN
                    IWARN = 2
                ELSE
                    IWARN = 1
                ENDIF
                F(N-1) = 0.0
                F(N) = 0.0
                GOTO 900
            ELSE
                LM = EIGVL(1,N-1)*EIGVL(1,N) - EIGVL(2,N-1)*EIGVL(2,N)
                S1 = EIGVL(1,N-1)+EIGVL(1,N) - A(N,N)
                S2 = (S1*A(N,N) - LM)/A(N,N-1)
                F(N-1) = (A(N-1,N-1) - S1)/BINP(N-1)
                F(N) = (A(N-1,N) - S2)/BINP(N-1)
                IEIGAL = IEIGAL + 2
            ENDIF
        ENDIF


900     CONTINUE

C
C.......apply accumulated reflectors to f vector
C
                
        DO 300 K=IREFC,1,-1
                CALL SPREM(F,LDA,IWORK(K),1,1,RWORK(IWRKDX+(K-1)*4),3)
300     CONTINUE

C
C.......reverse transformation to controllability form for
C       vector f
C

        IF(.NOT. IFORM) THEN
                II=N*(N+1)/2 + N - 4
                DO 350 K=2,N
                        CALL SPOST(F,1,N-K+1,N-K+1,1,RWORK(II),K)
                        II = II - K - 2
350             CONTINUE
        ENDIF

        RETURN
        END


C*********************************************************************
C
C       Title:          SCNTMK
C
C       Purpose:        Converts the system (b,A) into 
C                       controllability form.
C
C
C       Input Arguments:
C       ----------------
C
C       A       real  (lda,n)           Control matrix A
C                                               
C       n       integer                 Order of matrix A
C
C       lda     integer                 First dimension of array A as
C                                       defined in the calling program
C
C       binp    real  (n)               Input vector
C
C       cstor   real  (*)               Working storage for
C                                       conversion to controllability
C                                       form.
C
C
C
C       Local Variables:
C       ----------------
C
C       ipnt    integer                 Pointer to current reflector
C                                       in CSTOR()
C
C       
C       Subroutines Used:
C       -----------------
C
C       SCOPY, SNRM2, SVHOUS, SPREM, SPOST
C
C*********************************************************************

        SUBROUTINE SCNTMK(A,N,LDA,BINP,CSTOR)

        INTEGER N,LDA,IPNT
        REAL A(LDA,*),BINP(N),CSTOR(*),SNRM2


        IPNT = 1
        DO 100 I=N,2,-1

            IF(I .EQ. N) THEN
                CALL SCOPY(I, BINP(1), 1, CSTOR(IPNT+1), 1)
                BINP(1) = -SIGN(SNRM2(I, BINP(1), 1),BINP(1))
                DO 150 J=2,N
                    BINP(J) = 0.0
150             CONTINUE
            ELSE
                CALL SCOPY(I, A(N-I+1,N-I), 1, CSTOR(IPNT+1), 1)
                A(N-I+1,N-I) = -SIGN(SNRM2(I, A(N-I+1,N-I), 1),
     +                  A(N-I+1,N-I))
                DO 175 J=N-I+2,N
                    A(J,N-I) = 0.0
175             CONTINUE
            ENDIF

            CALL SVHOUS(CSTOR(IPNT), I)
            CALL SPREM(A,LDA,N-I+1,N-I+1,N,CSTOR(IPNT),I)
            CALL SPOST(A,LDA,1,N,N-I+1,CSTOR(IPNT),I)

            IPNT = IPNT + I + 1

100     CONTINUE


        RETURN
        END



C*********************************************************************
C
C       Title:          SVHOUS
C
C       Purpose:        Given a n-element vector stored in v(2)-
C                       v(n+1),  HOUSE calculates pi and u such that
C                       I - u * uT / pi will introduce zeroes into
C                       first two elements of v.  Pi is returned in
C                       v(1), while u overwrites the input vector
C                       v(2)-v(n+1).
C                       
C       Input Arguments:
C       ----------------
C
C       v       real  (*)       The last n elements of v contain
C                               the input vector to be zeroed.
C
C
C       Output Values:
C       --------------
C
C       v       v(1) = pi
C               v(2)-v(n+1) = u
C
C
C       Subroutines used:
C       -----------------
C
C       SSCAL, SNRM2, ISAMAX
C
C*********************************************************************

C
C.......overwrites v with u of elem. reflector; returns p
C
        SUBROUTINE SVHOUS(V,N)

        INTEGER N,I
        REAL V(*),SIGMA,ALFA,SNRM2

        I = ISAMAX( N, V(2), 1)
        ALFA = 1.0/ABS(V(I+1))
        CALL SSCAL( N, ALFA, V(2), 1)
        SIGMA = SIGN(SNRM2(N, V(2), 1),V(2))
        V(2) = V(2) + SIGMA
        V(1) = SIGMA*V(2)


        RETURN
        END



C*********************************************************************
C
C       Title:          SEIGCK
C
C       Purpose:        Checks that complex eigenvalues are
C                       complex conjugates.  The function
C                       returns .TRUE. if the eigenvalues are
C                       real or complex conjugates, .FALSE.
C                       otherwise.
C                       
C
C
C       Input Arguments:
C       ----------------
C
C       eigvl           real  (2,n)     Array of eigenvalues to
C                                       be allocated.
C
C       le              integer         Position of leading edge of
C                                       matrix A in array A.
C
C       tol             real            Epsilon value for
C                                       comparison to zero.
C
C                                       tol = n||(b,A)||u
C
C                                       where ||.|| is the
C                                       1-norm of the system,
C                                       and u is the machine unit.
C
C       Local Variables:
C       ----------------
C
C       rd1             real            Relative difference in real
C                                       parts.
C
C       rd2             real            Relative difference in complex
C                                       parts.
C
C       norm,tnorm,tol
C
C       
C       Subroutines Used:
C       -----------------
C
C       SASUM
C
C*********************************************************************

        LOGICAL FUNCTION SEIGCK(EIGVL,LE,TOL)

        REAL EIGVL(2,*),TOL,RD1,RD2
        INTEGER LE




        SEIGCK = .TRUE.

C
C.......check if complex eigenvalues are complex conjugates
C

        IF(ABS(EIGVL(2,LE)) .GT. TOL .OR. 
     +                          ABS(EIGVL(2,LE+1)) .GT. TOL) THEN
                RD1 = ABS((EIGVL(1,LE) - EIGVL(1,LE+1))/EIGVL(1,LE+1))
                RD2 = ABS((EIGVL(2,LE) + EIGVL(2,LE+1))/EIGVL(2,LE+1))
                IF(RD1 .GT. TOL .OR. RD2 .GT. TOL) THEN
                        SEIGCK = .FALSE.
                ENDIF
        ENDIF
                

        RETURN
        END


C*********************************************************************
C
C       Title:          SGTTOL
C
C       Purpose:        Calculates the tolerance used in determining
C                       possible uncontrollability of the system.
C                       If the input parameter TOL is positive, that
C                       value is used as the tolerance. If TOL is
C                       identically 0.0, the routine calculates the
C                       tolerance as follows:
C                       
C                       tol = n*||(b,A)||*u
C
C                       where ||.|| is the 1-norm, and u is the
C                       machine unit.
C
C                       A negative value for TOL causes the subroutine
C                       to halt and generates an error message.
C
C
C       Input Arguments:
C       ----------------
C
C       A       real  (lda,n)           Control matrix A
C                                               
C       n       integer                 Order of matrix A
C
C       lda     integer                 First dimension of array A as
C                                       defined in the calling program
C
C       b       real  (n)               Input vector b
C
C       tol     real                    User value for tolerance.
C
C       ierr    integer                 Error flag.  ierr = 0 means
C                                       no error.
C
C       Local Variables:
C       ----------------
C
C       norm,tnorm,eps
C
C       
C       Subroutines Used:
C       -----------------
C
C       SASUM, SSTORE
C
C*********************************************************************

        REAL FUNCTION SGTTOL(A,N,LDA,B,TOL,IERR)

        INTEGER N,LDA,IERR
        REAL A(LDA,*),B(N),TOL

        REAL NORM,TNORM,EPS,SASUM,EEPS

        COMMON/MACEPS/EEPS

C
C.......use user value for tolerance if non-zero
C
        IF(TOL .LT. 0.0) THEN
                IERR = 1
                RETURN
        ELSEIF(TOL .GT. 0.0) THEN
                SGTTOL = TOL
        ELSE

C.......calculate machine unit

                EPS = 1.0
30              EPS = EPS / 2.0
                EEPS = EPS + 1.0
                CALL SSTORE(EEPS)
                IF(EEPS .NE. 1.0) GOTO 30
                EPS = EPS * 2.0
C
C...............calculate 1 norm of system (b,A)
C
                NORM = SASUM(N, B(1), 1)
                DO 50 I=1,N
                        TNORM = SASUM(N, A(1,I), 1)
                        IF(TNORM .GT. NORM) NORM = TNORM
50              CONTINUE

                SGTTOL = N * NORM * EPS

        ENDIF


        RETURN
        END

C
C.......This subroutine ensures that the value E is actually
C       stored in memory, with appropriate rounding errors.
C       (From Gentleman and Marovich, Communications of the ACM,
C       vol. 17, no. 5, May 1974. p. 277)
C

        SUBROUTINE SSTORE(E)

        REAL E,V

        COMMON/MACEPS/V

        V = E

        RETURN
        END


C*********************************************************************
C
C       Title:          SCRH
C
C       Purpose:        Uses a series of elementary reflectors
C                       to restore the matrix A to a "near" upper
C                       Hessenberg form.
C
C
C       Input Arguments:
C       ----------------
C
C       A       real  (lda,n)           Control matrix A
C                                               
C       le      integer                 Position of leading edge of
C                                       matrix A in array A.
C
C       n       integer                 Order of matrix A
C
C       lda     integer                 First dimension of array A as
C                                       defined in the calling program
C
C       rstor   real  (4,*)             Working storage for 
C                                       reflectors
C
C       irpos   integer(*)              Position of reflectors
C
C       irefc   integer                 Index of last elementary
C                                       reflector stored in rstor.
C
C
C       Output Values:
C       --------------
C
C       A                       A is returned, transformed into near 
C                               upper Hessenberg.
C
C       rstor,irpos,irefc       Contain reflectors used to transform
C                               A to near upper Hessenberg.
C
C       Local Variables:
C       ----------------
C
C
C       
C       Subroutines Used:
C       -----------------
C
C       SCOPY, HOUSE, SNRM2, SPOST, SPREM
C
C*********************************************************************

        SUBROUTINE SCRH(A,LE,N,LDA,RSTOR,IRPOS,IREFC)

        INTEGER LE,N,LDA,I,IREFC,IRPOS(*)
        REAL A(LDA,*),P,RSTOR(4,*),SNRM2

        DO 100 I=N,LE+3,-1
                IREFC = IREFC + 1
                IRPOS(IREFC) = I-3
                CALL SCOPY(3, A(I,I-3), LDA, RSTOR(2,IREFC), 1)
                CALL HOUSE(RSTOR(1,IREFC))

                P = -SIGN(SNRM2(3, A(I,I-3), LDA),A(I,I-1))
                CALL SPOST(A,LDA,LE,I-1,I-3,RSTOR(1,IREFC),3)
                A(I,I-1) = P
                A(I,I-2) = 0.0
                A(I,I-3) = 0.0

                CALL SPREM(A,LDA,I-3,LE,N,RSTOR(1,IREFC),3)

100     CONTINUE

        RETURN
        END




C*********************************************************************
C
C       Title:          SAPPSH
C
C       Purpose:        Calculates an elementary reflector
C                       containing the implicit shift and pre-
C                       and post multiplies A.
C                       
C       Input Arguments:
C       ----------------
C
C       A       real  (lda,n)           Control matrix A.
C
C       le      integer                 Position of leading edge of
C                                       matrix A in array A.
C
C       n       integer                 Order of matrix A
C
C       lda     integer                 First dimension of array A as
C                                       defined in the calling program
C
C       eigvl   real  (2,n)             Vector of eigenvalues to be
C                                       allocated.
C
C       rstor   real  (4,*)             Working storage for reflectors.
C
C       irpos   integer(*)              Position of reflectors.
C
C       irefc   integer                 Index of last reflector in
C                                       irpos and rstor.
C
C
C       Output Values:
C       --------------
C
C       A               Returns the original matrix A pre- and 
C                       post-multiplied by the elementary reflector.
C
C       rstor,irpos     Stores the value and position of the 
C                       elementary reflector.
C
C       la,lm
C
C       Subroutines used:
C       -----------------
C
C       HOUSE, SPREM, SPOST
C
C*********************************************************************

        SUBROUTINE SAPPSH(A,LE,N,LDA,EIGVL,RSTOR,IRPOS,IREFC)

        INTEGER LE,N,LDA,IRPOS(*),IREFC
        REAL A(LDA,*),RSTOR(4,*)
        REAL EIGVL(2,*)

        REAL LA,LM

        IREFC = IREFC + 1
        IRPOS(IREFC) = N-2
C
C.......construct bottom row of A~
C
        LA = EIGVL(1,LE) + EIGVL(1,LE+1)
        LM = EIGVL(1,LE)*EIGVL(1,LE+1) - EIGVL(2,LE)*EIGVL(2,LE+1)
        RSTOR(2,IREFC) = A(N,N-1)*A(N-1,N-2)
        RSTOR(3,IREFC) = A(N,N-1)*A(N-1,N-1) + A(N,N)*A(N,N-1) - 
     +          LA*A(N,N-1)
        RSTOR(4,IREFC) = A(N,N-1)*A(N-1,N) + A(N,N)*A(N,N) -
     +          LA*A(N,N) + LM


C
C.......convert A to PtAP
C
        
        CALL HOUSE(RSTOR(1,IREFC))

        CALL SPOST(A,LDA,LE,N,N-2,RSTOR(1,IREFC),3)
        CALL SPREM(A,LDA,N-2,LE,N,RSTOR(1,IREFC),3)


        RETURN
        END



C*********************************************************************
C
C       Title:          SPREM
C
C       Purpose:        Premultiplies the matrix contained in array A
C                       by the 3-element elementary reflector stored 
C                       in vector v.
C                       
C       Input Arguments:
C       ----------------
C
C       A       real  (lda,nv)          Control matrix A.
C
C       lda     integer                 First dimension of array A as
C                                       defined in the calling program
C
C       irow    integer                 First row of A affected by
C                                       the reflector. (Reflector will
C                                       be applied to rows *irow* to
C                                       *irow+nv-1*.)
C
C       j,k     integer                 Range of columns of A on
C                                       which to apply reflector.
C
C       v       real  (4)               Contains the elementary 
C                                       reflector.  v(1) = pi and
C                                       v(2)->v(nv+1) = u, where
C                                       I - u*uT/pi is the reflector.
C
C       nv      integer                 Dimension of vector v.  
C
C
C       Output Values:
C       --------------
C
C       A               Returns the original matrix A pre-
C                       multiplied by the elementary reflector.
C
C
C
C       Subroutines used:
C       -----------------
C
C       SDOT, SAXPY
C
C*********************************************************************

        SUBROUTINE SPREM(A,LDA,IROW,J,K,V,NV)

        INTEGER IROW,J,K,LDA,I,NV
        REAL A(LDA,*),V(*),ALPHA,SDOT

        DO 100 I=J,K
        
                ALPHA = SDOT(NV, A(IROW,I), 1, V(2), 1)/V(1)
                CALL SAXPY(NV, -ALPHA, V(2), 1, A(IROW,I), 1)

100     CONTINUE



        RETURN
        END


C*********************************************************************
C
C       Title:          SPOST
C
C       Purpose:        post-multiplies the matrix contained in 
C                       array A by the 3-element elementary reflector 
C                       stored in vector v.
C                       
C       Input Arguments:
C       ----------------
C
C       A       real  (lda,n)           Control matrix A.
C
C       lda     integer                 First dimension of array A as
C                                       defined in the calling program
C
C       j,k     integer                 Range of rows of A on
C                                       which to apply reflector.
C
C       icol    integer                 First column of A affected by
C                                       the reflector. (Reflector will
C                                       be applied to columns *icol* 
C                                       to *icol+nv-1*.)
C
C       v       real  (4)               Contains the elementary 
C                                       reflector.  v(1) = pi and
C                                       v(2)->v(nv+1) = u, where
C                                       I - u*uT/pi is the reflector.
C
C       nv      integer                 Dimension of vector v.
C                                       
C
C       Output Values:
C       --------------
C
C       A               Returns the original matrix A 
C                       post-multiplied by the elementary reflector.
C
C
C
C       Subroutines used:
C       -----------------
C
C       SDOT, SAXPY
C
C*********************************************************************

        SUBROUTINE SPOST(A,LDA,J,K,ICOL,V,NV)

        INTEGER J,K,ICOL,LDA,I,NV
        REAL A(LDA,*),V(*),ALPHA,SDOT

        DO 100 I=J,K
        
                ALPHA = SDOT(NV, A(I,ICOL), LDA, V(2), 1)/V(1)
                CALL SAXPY(NV, -ALPHA, V(2), 1, A(I,ICOL), LDA)

100     CONTINUE



        RETURN
        END


C*********************************************************************
C
C       Title:          HOUSE
C
C       Purpose:        Given a 3-element vector stored in v(2)-
C                       v(4),  HOUSE calculates pi and u such that
C                       I - u * uT / pi will introduce zeroes into
C                       first two elements of v.  Pi is returned in
C                       v(1), while u overwrites the input vector
C                       v(2)-v(4).
C                       
C       Input Arguments:
C       ----------------
C
C       v       real  (4)       The last three elements of v contain
C                               the input vector to be zeroed.
C
C
C       Output Values:
C       --------------
C
C       v       v(1) = pi
C               v(2)-v(4) = u
C
C
C       Subroutines used:
C       -----------------
C
C       SSCAL, SNRM2, ISAMAX
C
C*********************************************************************

C
C.......overwrites v with u of elem. reflector; returns p
C
        SUBROUTINE HOUSE(V)

        INTEGER I
        REAL V(4),SIGMA,ALFA,SNRM2

        I = ISAMAX( 3, V(2), 1)
        ALFA = 1.0/ABS(V(I+1))
        CALL SSCAL( 3, ALFA, V(2), 1)
        SIGMA = SIGN(SNRM2(3, V(2), 1),V(4))
        V(4) = V(4) + SIGMA
        V(1) = SIGMA*V(4)


        RETURN
C *** Last line of SSEVAS.F ***
        END
